from functools import partial

from visidata import VisiData, vd, Sheet, date, anytype, Path, options, Column, asyncthread, Progress, undoAttrCopyFunc, run

@VisiData.api
def open_pandas(vd, p):
    return PandasSheet(p.base_stem, source=p)

@VisiData.api
def open_dta(vd, p):
    return PandasSheet(p.base_stem, source=p, filetype='stata')

VisiData.open_stata = VisiData.open_pandas

for ft in 'feather gbq orc pickle sas stata'.split():
    funcname ='open_'+ft
    if not getattr(VisiData, funcname, None):
        setattr(VisiData, funcname, lambda vd,p,ft=ft: PandasSheet(p.base_stem, source=p, filetype=ft))

@VisiData.api
@asyncthread
def save_dta(vd, p, *sheets):
    import pandas as pd
    import numpy as np

    # STATA is a one-sheet software
    # Save only the first sheet
    vs = sheets[0]

    columns = [col.name for col in vs.visibleCols]

    # Get data types
    types = list()
    dispvals = next(vs.iterdispvals(format=True))
    for col,_ in dispvals.items():
        if col.type in [bool, int, float]:
            types.append(col.type)
        elif vd.isNumeric(col):
            types.append(float)
        else:
            types.append(str)

    # Populate numpy array
    data = np.empty((vs.nRows, len(columns)), dtype=object)
    for r_i, dispvals in enumerate(vs.iterdispvals(format=True)):
        for c_i, v in enumerate(dispvals.values()):
            data[r_i, c_i] = v

    # Convert to pandas DataFrame and save
    dtype = {col:t for col,t in zip(columns, types)}
    df = pd.DataFrame(data, columns=columns)
    df = df.astype(dtype)
    df.to_stata(p, version=118, write_index=False)

class DataFrameAdapter:
    def __init__(self, df):
        pd = vd.importExternal('pandas')
        if not isinstance(df, pd.DataFrame):
            vd.fail('%s is not a dataframe' % type(df).__name__)

        self.df = df

    def __len__(self):
        if 'df' not in self.__dict__:
            return 0
        return len(self.df)

    def __getitem__(self, k):
        if isinstance(k, slice):
            return DataFrameAdapter(self.df.iloc[k])
        return self.df.iloc[k]

    def __getattr__(self, k):
        if 'df' not in self.__dict__:
            raise AttributeError(f"'{self.__class__.__name__}' has no attribute '{k}'")
        return getattr(self.df, k)

# source=DataFrame
class PandasSheet(Sheet):
    '''Sheet sourced from a pandas.DataFrame

    Warning:
        The index of the pandas.DataFrame input must be unique.
        Otherwise the selection functionality, which relies on
        looking up selected rows via the index, will break.
        This can be done by calling reset_index().

    Note:
        Columns starting with "__vd_" are reserved for internal usage
        by the VisiData loader.
    '''

    def dtype_to_type(self, dtype):
        np = vd.importExternal('numpy')
        # Find the underlying numpy dtype for any pandas extension dtypes
        dtype = getattr(dtype, 'numpy_dtype', dtype)
        try:
            if np.issubdtype(dtype, np.integer):
                return int
            if np.issubdtype(dtype, np.floating):
                return float
            if np.issubdtype(dtype, np.datetime64):
                return date
        except TypeError:
            # For categoricals and other pandas-defined dtypes
            pass
        return anytype

    def read_tsv(self, path, **kwargs):
        'Partial function for reading TSV files using pd.read_csv'
        pd = vd.importExternal('pandas')
        return pd.read_csv(path, sep='\t', **kwargs)

    @property
    def df(self):
        if isinstance(getattr(self, 'rows', None), DataFrameAdapter):
            return self.rows.df

    @df.setter
    def df(self, val):
        if isinstance(getattr(self, 'rows', None), DataFrameAdapter):
            self.rows.df = val
        else:
            self.rows = DataFrameAdapter(val)

    def getValue(self, col, row):
        '''Look up column values in the underlying DataFrame.'''
        return col.sheet.df.loc[row.name, col.expr]

    def setValue(self, col, row, val):
        '''
        Update a column's value in the underlying DataFrame, loosening the
        column's type as needed. Take care to avoid assigning to a view or
        a copy as noted here:

        https://pandas.pydata.org/pandas-docs/stable/user_guide/indexing.html#why-does-assignment-fail-when-using-chained-indexing
        '''
        try:
            col.sheet.df.loc[row.name, col.expr] = val
        except ValueError as err:
            vd.warning(f'Type of {val} does not match column {col.name}. Changing type.')
            col.type = anytype
            col.sheet.df.loc[row.name, col.expr] = val
        self.setModified()

    @asyncthread
    def reload(self):
        pd = vd.importExternal('pandas')
        if isinstance(self.source, pd.DataFrame):
            df = self.source
        elif isinstance(self.source, Path):
            filetype = getattr(self, 'filetype', self.source.ext)
            if filetype == 'tsv':
                readfunc = self.read_tsv
            elif filetype == 'jsonl':
                readfunc = partial(pd.read_json, lines=True)
            elif filetype == 'hdf5':
                readfunc = partial(pd.read_hdf, lines=True)
            else:
                readfunc = getattr(pd, 'read_'+filetype) or vd.error('no pandas.read_'+filetype)
            # readfunc() handles binary and text open()
            df = readfunc(self.source, **options.getall('pandas_'+filetype+'_'))
            # some read methods (html, for example) return a list of dataframes
            if isinstance(df, list):
                for idx, inner_df in enumerate(df[1:], start=1):
                    vd.push(PandasSheet(f'{self.name}[{idx}]', source=inner_df))
                df = df[0]
                self.name += '[0]'
            if (filetype == 'pickle') and not isinstance(df, pd.DataFrame):
                vd.fail('pandas loader can only unpickle dataframes')
        else:
            try:
                df = pd.DataFrame(self.source)
            except ValueError as err:
                vd.fail('error building pandas DataFrame from source data: %s' % err)

        # reset the index here
        if type(df.index) is not pd.RangeIndex:
            df = df.reset_index(drop=True)

        # VisiData assumes string column names but pandas does not. Forcing string
        # columns at load-time avoids various errors later.
        df.columns = df.columns.astype(str)

        self.columns = []
        for col in (c for c in df.columns if not c.startswith("__vd_")):
            self.addColumn(Column(
                col,
                type=self.dtype_to_type(df[col]),
                getter=self.getValue,
                setter=self.setValue,
                expr=col
            ))

        if self.columns[0].name == 'index': # if the df contains an index column
            self.column('index').hide()

        self.rows = DataFrameAdapter(df)
        self._selectedMask = pd.Series(False, index=df.index)
        if df.index.nunique() != df.shape[0]:
            vd.warning("Non-unique index, row selection API may not work or may be incorrect")

    @asyncthread
    def sort(self):
        '''Sort rows according to the current self._ordering.'''
        by_cols = []
        ascending = []
        for col, reverse in self._ordering[::-1]:
            by_cols.append(col.expr)
            ascending.append(not reverse)
        self.rows.sort_values(by=by_cols, ascending=ascending, inplace=True)

    def _checkSelectedIndex(self):
        pd = vd.importExternal('pandas')
        if self._selectedMask.index is not self.df.index:
            # DataFrame was modified inplace, so the selection is no longer valid
            vd.status('pd.DataFrame.index updated, clearing {} selected rows'
                      .format(self._selectedMask.sum()))
            self._selectedMask = pd.Series(False, index=self.df.index)

    def rowid(self, row):
        return getattr(row, 'name', None) or ''

    # Base selection API. Refer to GH #266: using id() will not identify
    # pandas rows since iterating on rows / selecting rows will return
    # different copies. Instead, re-implement the selection API by
    # keeping a boolean pd.Series indicating the selected rows.
    def isSelected(self, row):
        if row is None:
            return False
        self._checkSelectedIndex()
        return self._selectedMask.loc[row.name]

    def selectRow(self, row):
        'Select given row'
        self._checkSelectedIndex()
        self._selectedMask.loc[row.name] = True

    def unselectRow(self, row):
        self._checkSelectedIndex()
        is_selected = self._selectedMask.loc[row.name]
        self._selectedMask.loc[row.name] = False
        return is_selected

    @property
    def nSelectedRows(self):
        self._checkSelectedIndex()
        return self._selectedMask.sum()

    @property
    def selectedRows(self):
        self._checkSelectedIndex()
        return DataFrameAdapter(self.df.loc[self._selectedMask])

    # Vectorized implementation of multi-row selections
    @asyncthread
    def select(self, rows, status=True, progress=True):
        self.addUndoSelection()
        for row in (Progress(rows, 'selecting') if progress else rows):
            self.selectRow(row)

    @asyncthread
    def unselect(self, rows, status=True, progress=True):
        self.addUndoSelection()
        for row in (Progress(rows, 'unselecting') if progress else rows):
            self.unselectRow(row)

    def clearSelected(self):
        pd = vd.importExternal('pandas')
        self._selectedMask = pd.Series(False, index=self.df.index)

    def selectByIndex(self, start=None, end=None):
        self._checkSelectedIndex()
        self._selectedMask.iloc[start:end] = True

    def unselectByIndex(self, start=None, end=None):
        self._checkSelectedIndex()
        self._selectedMask.iloc[start:end] = False

    def toggleByIndex(self, start=None, end=None):
        self._checkSelectedIndex()
        self.addUndoSelection()
        self._selectedMask.iloc[start:end] = ~self._selectedMask.iloc[start:end]

    def _selectByILoc(self, mask, selected=True):
        self._checkSelectedIndex()
        self._selectedMask.iloc[mask] = selected

    @asyncthread
    def selectByRegex(self, regex, columns, unselect=False):
        '''
        Find rows matching regex in the provided columns. By default, add
        matching rows to the selection. If unselect is True, remove from the
        active selection instead.
        '''
        pd = vd.importExternal('pandas')
        case_sensitive = 'I' not in vd.options.regex_flags
        masks = pd.DataFrame([
            self.df[col.expr].astype(str).str.contains(pat=regex, case=case_sensitive, regex=True)
            for col in columns
        ])
        if unselect:
            self._selectedMask = self._selectedMask & ~masks.any()
        else:
            self._selectedMask = self._selectedMask | masks.any()

    def addUndoSelection(self):
        vd.addUndo(undoAttrCopyFunc([self], '_selectedMask'))

    @property
    def nRows(self):
        if self.df is None:
            return 0
        return len(self.df)

    def newRows(self, n):
        '''
        Return n rows of empty data. Let pandas decide on the most
        appropriate missing value (NaN, NA, etc) based on the underlying
        DataFrame's dtypes.
        '''

        pd = vd.importExternal('pandas')
        return pd.DataFrame({
            col: [None] * n for col in self.df.columns
        }).astype(self.df.dtypes.to_dict(), errors='ignore')

    def addRows(self, rows, index=None, undo=True):
        pd = vd.importExternal('pandas')
        if index is None:
            self.df = self.df.append(pd.DataFrame(rows))
        else:
            self.df = pd.concat((self.df.iloc[0:index], pd.DataFrame(rows), self.df.iloc[index:]))
        self.df.index = pd.RangeIndex(self.nRows)
        self._checkSelectedIndex()
        if undo:
            self.setModified()
            vd.addUndo(self._deleteRows, range(index, index + len(rows)))

    def _deleteRows(self, which):
        pd = vd.importExternal('pandas')
        self.df.drop(which, inplace=True)
        self.df.index = pd.RangeIndex(self.nRows)
        self._checkSelectedIndex()

    def addRow(self, row, index=None):
        self.addRows([row], index)
        vd.addUndo(self._deleteRows, index or self.nRows - 1)

    def delete_row(self, rowidx):
        pd = vd.importExternal('pandas')
        oldrow = self.df.iloc[rowidx:rowidx+1]

        # Use to_dict() here to work around an edge case when applying undos.
        # As an action is undone, its entry gets removed from the cmdlog sheet.
        # If we use `oldrow` directly, we get errors comparing DataFrame objects
        # when there are multiple deletion commands for the same row index.
        # There may be a better way to handle that case.
        vd.addUndo(self.addRows, oldrow.to_dict(), rowidx, undo=False)
        self._deleteRows(rowidx)
        vd.memory.cliprows = [oldrow]
        self.setModified()

    def deleteBy(self, by):
        '''Delete rows for which func(row) is true.  Returns number of deleted rows.'''
        pd = vd.importExternal('pandas')
        nRows = self.nRows
        vd.addUndo(setattr, self, 'df', self.df.copy())
        self.df = self.df[~by]
        self.df.index = pd.RangeIndex(self.nRows)
        ndeleted = nRows - self.nRows

        self.setModified()
        vd.status('deleted %s %s' % (ndeleted, self.rowtype))
        return ndeleted

    def deleteSelected(self):
        '''Delete all selected rows.'''
        self.deleteBy(self._selectedMask)


@VisiData.global_api
def view_pandas(vd, df):
    run(PandasSheet('', source=df))


# Override with vectorized implementations
PandasSheet.addCommand(None, 'stoggle-rows', 'toggleByIndex()', 'toggle selection of all rows')
PandasSheet.addCommand(None, 'select-rows', 'selectByIndex()', 'select all rows')
PandasSheet.addCommand(None, 'unselect-rows', 'unselectByIndex()', 'unselect all rows')

PandasSheet.addCommand(None, 'stoggle-before', 'toggleByIndex(end=cursorRowIndex)', 'toggle selection of rows from top to cursor')
PandasSheet.addCommand(None, 'select-before', 'selectByIndex(end=cursorRowIndex)', 'select all rows from top to cursor')
PandasSheet.addCommand(None, 'unselect-before', 'unselectByIndex(end=cursorRowIndex)', 'unselect all rows from top to cursor')
PandasSheet.addCommand(None, 'stoggle-after', 'toggleByIndex(start=cursorRowIndex)', 'toggle selection of rows from cursor to bottom')
PandasSheet.addCommand(None, 'select-after', 'selectByIndex(start=cursorRowIndex)', 'select all rows from cursor to bottom')
PandasSheet.addCommand(None, 'unselect-after', 'unselectByIndex(start=cursorRowIndex)', 'unselect all rows from cursor to bottom')
PandasSheet.addCommand(None, 'random-rows', 'nrows=int(input("random number to select: ", value=nRows)); vs=copy(sheet); vs.name=name+"_sample"; vs.rows=DataFrameAdapter(sheet.df.sample(nrows or nRows)); vd.push(vs)', 'open duplicate sheet with a random population subset of N rows')

# Handle the regex selection family of commands through a single method,
# since the core logic is shared
PandasSheet.addCommand('|', 'select-col-regex', 'selectByRegex(regex=inputRegex("select regex: ", defaultLast=True), columns=[cursorCol])', 'select rows matching regex in current column')
PandasSheet.addCommand('\\', 'unselect-col-regex', 'selectByRegex(regex=inputRegex("select regex: ", defaultLast=True), columns=[cursorCol], unselect=True)', 'unselect rows matching regex in current column')
PandasSheet.addCommand('g|', 'select-cols-regex', 'selectByRegex(regex=inputRegex("select regex: ", defaultLast=True), columns=visibleCols)', 'select rows matching regex in any visible column')
PandasSheet.addCommand('g\\', 'unselect-cols-regex', 'selectByRegex(regex=inputRegex("select regex: ", defaultLast=True), columns=visibleCols, unselect=True)', 'unselect rows matching regex in any visible column')

# Override with a pandas/dataframe-aware implementation
PandasSheet.addCommand('"', 'dup-selected', 'vs=PandasSheet(sheet.name, "selectedref", source=selectedRows.df); vd.push(vs)', 'open duplicate sheet with only selected rows')

vd.addGlobals({
    'PandasSheet': PandasSheet,
})



from visidata import vd, Column, VisiData, ItemColumn, Path, AttrDict, BaseSheet, IndexSheet
from visidata import RowColorizer, CellColorizer
from visidata import filesize, modtime, date

from .gitsheet import GitSheet
#from .diff import DifferSheet

vd.option('vgit_show_ignored', False, 'show ignored files on git status')
vd.theme_option('color_git_staged_mod', 'green', 'color of files staged with modifications')
vd.theme_option('color_git_staged_add', 'green', 'color of files staged for addition')
vd.theme_option('color_git_staged_del', 'red', 'color of files staged for deletion')
vd.theme_option('color_git_unstaged_del', 'on 88', 'color of files deleted but unstaged')
vd.theme_option('color_git_untracked', '243 blue', 'color of ignored/untracked files')


@VisiData.api
def git_status(vd, p, args, **kwargs):
    vs = GitStatus('/'.join(p.parts[-2:]), source=p)
    if not vs.gitRootPath:
        return vd.git_repos(p, [])
    return vs

class GitFile:
    def __init__(self, path, gitsrc):
        self.path = path
        self.filename = path.relative_to(gitsrc)
        self.is_dir = self.path.is_dir()

    def __str__(self):
        return str(self.filename) + (self.is_dir and '/' or '')

class GitStatus(GitSheet):
    rowtype = 'files'  # rowdef: GitFile
    guide = '''
        # git status
        An overview of the local git checkout.

        - `Enter` to open diff of file (`git diff`)
        - `a` to stage changes in file (`git add`)
        - `r` to unstage changes in file (`git reset`)
        - `c` to revert all unstaged changes in file (`git checkout`)
        - `d` to stage the entire file for deletion (`git rm`)
        - `z Ctrl+S` to commit staged changes (`git commit`)
    '''

    columns = [
        Column('path', width=40, getter=lambda c,r: str(r)),
        Column('status', getter=lambda c,r: c.sheet.statusText(c.sheet.git_status(r)), width=8),
        Column('status_raw', getter=lambda c,r: c.sheet.git_status(r), width=0),
        Column('staged', getter=lambda c,r: c.sheet.git_status(r).dels),
        Column('unstaged', getter=lambda c,r: c.sheet.git_status(r).adds),
        Column('type', getter=lambda c,r: r.is_dir() and '/' or r.suffix, width=0),
        Column('size', type=int, getter=lambda c,r: filesize(r)),
        Column('modtime', type=date, getter=lambda c,r: modtime(r)),
    ]
    nKeys = 1

    colorizers = [
        CellColorizer(3, 'color_git_staged_mod',   lambda s,c,r,v: r and c and c.name == 'staged' and s.git_status(r).status[0] == 'M'), # staged mod
        CellColorizer(1, 'color_git_staged_del',     lambda s,c,r,v: r and c and c.name == 'staged' and s.git_status(r).status == 'D '), # staged delete
        RowColorizer(1, 'color_git_staged_add',  lambda s,c,r,v: r and s.git_status(r).status in ['A ', 'M ']), # staged add/mod
        RowColorizer(1, 'color_git_unstaged_del',       lambda s,c,r,v: r and s.git_status(r).status[1] == 'D'), # unstaged delete
        RowColorizer(3, 'color_git_untracked', lambda s,c,r,v: r and s.git_status(r).status == '!!'),  # ignored
        RowColorizer(1, 'color_git_untracked', lambda s,c,r,v: r and s.git_status(r).status == '??'),  # untracked
    ]

    def statusText(self, st):
        vmod = {'A': 'add', 'D': 'rm', 'M': 'mod', 'T': 'chmod', '?': '', '!': 'ignored', 'U': 'unmerged'}
        x, y = st.status
        if st == '??': # untracked
            return 'new'
        elif st == '!!':  # ignored
            return 'ignored'
        elif x != ' ' and y == ' ': # staged
            return vmod.get(x, x)
        elif y != ' ': # unstaged
            return vmod.get(y, y)
        else:
            return ''

    @property
    def workdir(self):
        return str(self.source)

    def git_status(self, r):
        '''return tuple of (status, adds, dels).
        status like !! ??
        adds and dels are lists of additions and deletions.
        '''
        if not r:
            return None

        fn = str(r)
        ret = self._cachedStatus.get(fn, None)
        if not ret:
            ret = AttrDict(status='??')
            self._cachedStatus[fn] = ret

        return ret

    def ignored(self, fn):
        if self.options.vgit_show_ignored:
            return False

        if fn in self._cachedStatus:
            return self._cachedStatus[fn].status == '!!'

        return False

    @property
    def remotediff(self):
        return self.gitBranchStatuses.get(self.branch, 'no branch')

    def iterload(self):
        files = [GitFile(p, self.source) for p in self.source.iterdir() if p.stem not in ('.git')]  # files in working dir

        filenames = dict((gf.filename, gf) for gf in files)

        self._cachedStatus.clear()
        for fn in self.git_iter('ls-files', '-z'):
            self._cachedStatus[fn] = AttrDict(status='  ')

        for line in self.git_iter('status', '-z', '-unormal', '--ignored'):
            if not line: continue

            if line[2:3] == ' ':
                st, fn = line[:2], line[3:]
            else:
                fn = line
                st = '??'  # untracked file

            self._cachedStatus[fn] = AttrDict(status=st)
            if not self.ignored(fn):
                yield Path(fn)

        for line in self.git_iter('diff-files', '--numstat', '-z'):
            if not line: continue
            adds, dels, fn = line.split('\t')
            if fn not in self._cachedStatus:
                self._cachedStatus[fn] = AttrDict(status='##')
            cs = self._cachedStatus[fn]
            cs.adds = '+%s/-%s' % (adds, dels)

        for line in self.git_iter('diff-index', '--cached', '--numstat', '-z', 'HEAD'):
            if not line: continue
            adds, dels, fn = line.split('\t')
            if fn not in self._cachedStatus:
                self._cachedStatus[fn] = AttrDict(status='$$')
            cs = self._cachedStatus[fn]
            cs.dels = '+%s/-%s' % (adds, dels)

        self.orderBy(None, self.columns[-1], reverse=True)

        self.recalc()  # erase column caches

    def openRow(self, row):
        'Open unstaged diffs for this file, or dive into directory'
        if row.is_dir:
            return GitStatus(row.path)
        else:
            return DifferSheet(row, "HEAD", "index", "working", source=sheet)

    def openRows(self, rows):
        'Open unstaged hunks for selected rows'
        return getHunksSheet(sheet, *rows)


@GitStatus.lazy_property
def _cachedStatus(self):
    return {}  # [filename] -> AttrDict(status='xx', adds=, dels=)


GitStatus.addCommand('a', 'git-add', 'loggit("add", cursorRow.filename)', 'add this new file or modified file to staging')
#GitStatus.addCommand('m', 'git-mv', 'loggit("mv", cursorRow.filename, input("rename file to: ", value=cursorRow.filename))', 'rename this file')
GitStatus.addCommand('d', 'git-rm', 'loggit("rm", cursorRow.filename)', 'stage this file for deletion')
GitStatus.addCommand('r', 'git-reset', 'loggit("reset", "HEAD", cursorRow.filename)', 'reset/unstage this file')
GitStatus.addCommand('c', 'git-checkout', 'loggit("checkout", cursorRow.filename)', 'checkout this file')
GitStatus.addCommand('ga', 'git-add-selected', 'loggit("add", *[r for r in selectedRows])', 'add all selected files to staging')
GitStatus.addCommand('gd', 'git-rm-selected', 'loggit("rm", *[r for r in selectedRows])', 'delete all selected files')
GitStatus.addCommand(None, 'git-commit', 'loggit("commit", "-m", input("commit message: "))', 'commit changes')
GitStatus.addCommand(None, 'git-ignore-file', 'open(rootPath/".gitignore", "a").write(cursorRow.filename+"\\n"); reload()', 'add file to toplevel .gitignore')
GitStatus.addCommand(None, 'git-ignore-wildcard', 'open(rootPath/.gitignore, "a").write(input("add wildcard to .gitignore: "))', 'add input line to toplevel .gitignore')


#GitStatus.addCommand('z^J', 'diff-file-staged', 'vd.push(getStagedHunksSheet(sheet, cursorRow))', 'push staged diffs for this file')
#GitStatus.addCommand('gz^J', 'diff-selected-staged', 'vd.push(getStagedHunksSheet(sheet, *(selectedRows or rows)))', 'push staged diffs for selected files or all files')
#GitStatus.addCommand('^O', 'sysopen-row', 'launchExternalEditorPath(Path(cursorRow.path))', 'open this file in $EDITOR')


vd.addMenuItems('''
    Git > View staged changes > current file > diff-file-staged
    Git > View staged changes > selected files > staged changes > diff-selected-staged
    Git > Stage > current file > git-add
    Git > Stage > selected files > git-add-selected
    Git > Unstage > current file > git-reset
    Git > Unstage > selected files > git-reset-selected
    Git > Rename file > git-mv
    Git > Delete > file > git-rm
    Git > Delete > selected files > git-rm-selected
    Git > Ignore > file > ignore-file
    Git > Ignore > wildcard > ignore-wildcard
    Git > Commit staged changes > git-commit
    Git > Revert unstaged changes > current file > git-checkout
''')



'''slide rows/columns around'''

import visidata
from visidata import Sheet, moveListItem, vd

@Sheet.api
def slide_col(sheet, colidx, newcolidx):
    vd.addUndo(moveVisibleCol, sheet, newcolidx, colidx)
    return moveVisibleCol(sheet, colidx, newcolidx)

@Sheet.api
def slide_keycol(sheet, fromKeyColIdx, toKeyColIdx):
    vd.addUndo(moveKeyCol, sheet, toKeyColIdx, fromKeyColIdx)
    return moveKeyCol(sheet, fromKeyColIdx, toKeyColIdx)


@Sheet.api
def slide_row(sheet, rowidx, newcolidx):
    vd.addUndo(moveListItem, sheet.rows, newcolidx, rowidx)
    return moveListItem(sheet.rows, rowidx, newcolidx)


def moveKeyCol(sheet, fromKeyColIdx, toKeyColIdx):
    'Move key column to another key column position in sheet.'
    if not (1 <= toKeyColIdx <= len(sheet.keyCols)):
        vd.warning('already at edge')
        return fromKeyColIdx-1

    for col in sheet.keyCols:
        if col.keycol == fromKeyColIdx:
            col.keycol = toKeyColIdx
        elif toKeyColIdx < fromKeyColIdx:  # moving to the left
            if toKeyColIdx <= col.keycol < fromKeyColIdx:
                col.keycol += 1
        else:  # moving to the right
            if fromKeyColIdx < col.keycol <= toKeyColIdx:
                col.keycol -= 1

    # key columns are 1-indexed; columns in general are 0-indexed
    return toKeyColIdx-1


def moveVisibleCol(sheet, fromVisColIdx, toVisColIdx):
    'Move visible column to another visible index in sheet.'
    # a regular column cannot move to the left of keycols
    if 0 <= toVisColIdx < sheet.nVisibleCols:
        fromVisColIdx = min(max(fromVisColIdx, 0), sheet.nVisibleCols-1)
        fromColIdx = sheet.columns.index(sheet.visibleCols[fromVisColIdx])
        if toVisColIdx < len(sheet.keyCols):
            vd.warning('already at edge')
            return fromVisColIdx
        else:
            toColIdx = sheet.columns.index(sheet.visibleCols[toVisColIdx])
        moveListItem(sheet.columns, fromColIdx, toColIdx)
        return toVisColIdx
    else:
        vd.warning('already at edge')
        return fromVisColIdx


Sheet.addCommand('H', 'slide-left', 'sheet.cursorVisibleColIndex = slide_col(cursorVisibleColIndex, cursorVisibleColIndex-1) if not cursorCol.keycol else slide_keycol(cursorCol.keycol, cursorCol.keycol-1)', 'slide the current column left **one position**')
Sheet.addCommand('L', 'slide-right', 'sheet.cursorVisibleColIndex = slide_col(cursorVisibleColIndex, cursorVisibleColIndex+1) if not cursorCol.keycol else slide_keycol(cursorCol.keycol, cursorCol.keycol+1)', 'slide the current column right **one position**')
Sheet.addCommand('J', 'slide-down', 'sheet.cursorRowIndex = slide_row(cursorRowIndex, cursorRowIndex+1)', 'slide current row down')
Sheet.addCommand('K', 'slide-up', 'sheet.cursorRowIndex = slide_row(cursorRowIndex, cursorRowIndex-1)', 'slide current row up')
Sheet.addCommand('gH', 'slide-leftmost', 'slide_col(cursorVisibleColIndex, len(keyCols) + 0) if not cursorCol.keycol else slide_keycol(cursorCol.keycol, 1)', 'slide the current column **all the way** to the left of its section')
Sheet.addCommand('gL', 'slide-rightmost', 'slide_col(cursorVisibleColIndex, nVisibleCols-1) if not cursorCol.keycol else slide_keycol(cursorCol.keycol, len(keyCols))', 'slide the current column **all the way** to the right of its section')
Sheet.addCommand('gJ', 'slide-bottom', 'slide_row(cursorRowIndex, nRows)', 'slide current row all the way to the bottom of sheet')
Sheet.addCommand('gK', 'slide-top', 'slide_row(cursorRowIndex, 0)', 'slide current row to top of sheet')
Sheet.addCommand('zH', 'slide-left-n', 'slide_col(cursorVisibleColIndex, cursorVisibleColIndex-int(input("slide col left n=", value=1)))', 'slide current column **N positions** to the left')
Sheet.addCommand('zL', 'slide-right-n', 'slide_col(cursorVisibleColIndex, cursorVisibleColIndex+int(input("slide col left n=", value=1)))', 'slide current column **N positions** to the right')
Sheet.addCommand('zJ', 'slide-down-n', 'slide_row(cursorRowIndex, cursorRowIndex+int(input("slide row down n=", value=1)))', 'slide current row N positions down')
Sheet.addCommand('zK', 'slide-up-n', 'slide_row(cursorRowIndex, cursorRowIndex-int(input("slide row up n=", value=1)))', 'slide current row N positions up')

Sheet.bindkey('Shift+Left', 'slide-left')
Sheet.bindkey('Shift+Up', 'slide-up')
Sheet.bindkey('Shift+Down', 'slide-down')
Sheet.bindkey('Shift+Right', 'slide-right')

Sheet.bindkey('g Shift+Left', 'slide-leftmost')
Sheet.bindkey('g Shift+Down', 'slide-bottom')
Sheet.bindkey('g Shift+Up', 'slide-top')
Sheet.bindkey('g Shift+Right', 'slide-rightmost')

vd.addMenuItems('''
    Edit > Slide > Row > up > slide-up
    Edit > Slide > Row > up N > slide-up-n
    Edit > Slide > Row > down > slide-down
    Edit > Slide > Row > down N > slide-down-n
    Edit > Slide > Row > to top > slide-top
    Edit > Slide > Row > to bottom > slide-bottom
    Edit > Slide > Column > left > slide-left
    Edit > Slide > Column > left N > slide-left-n
    Edit > Slide > Column > leftmost > slide-leftmost
    Edit > Slide > Column > right > slide-right
    Edit > Slide > Column > right N > slide-right-n
    Edit > Slide > Column > rightmost > slide-rightmost
''')

## tests

sample_path = vd.pkg_resources_files(visidata) / 'tests/sample.tsv'
benchmark_path = vd.pkg_resources_files(visidata) / 'tests/benchmark.csv'

def make_tester(setup_vdx):
    def t(vdx, golden):
        global vd
        vd = visidata.vd.resetVisiData()
        vd.runvdx(setup_vdx)

        vd.runvdx(vdx)
        colnames = [c.name for c in vd.sheet.visibleCols]
        assert colnames == golden.split(), ' '.join(colnames)

    return t

def test_slide_keycol_1(vd):
    t = make_tester(f'''
            open-file {sample_path}
            +::OrderDate key-col
            +::Region key-col
            +::Rep key-col
        ''')

    t('',                             'OrderDate Region Rep Item Units Unit_Cost Total')
    t('+::Rep slide-leftmost',        'Rep OrderDate Region Item Units Unit_Cost Total')
    t('+::OrderDate slide-rightmost', 'Region Rep OrderDate Item Units Unit_Cost Total')
    t('+::Rep slide-left',            'OrderDate Rep Region Item Units Unit_Cost Total')
    t('+::OrderDate slide-right',     'Region OrderDate Rep Item Units Unit_Cost Total')

    t('''
        +::Item key-col
        +::Item slide-left
        slide-left
        slide-right
        slide-right
        slide-left
        slide-left
    ''', 'OrderDate Item Region Rep Units Unit_Cost Total')


def test_slide_leftmost(vd):
    t = make_tester(f'''open-file {benchmark_path}''')

    t('+::Paid slide-leftmost', 'Paid Date Customer SKU Item Quantity Unit')

    t = make_tester(f'''
         open-file {benchmark_path}
         +::Date key-col
    ''')

    t('', 'Date Customer SKU Item Quantity Unit Paid')
    t('''+::Item slide-leftmost''', 'Date Item Customer SKU Quantity Unit Paid')
    t('''+::SKU key-col
         +::Quantity slide-leftmost''', 'Date SKU Quantity Customer Item Unit Paid')
    t('''+::Date slide-leftmost''', 'Date Customer SKU Item Quantity Unit Paid')
    t('''+::Item slide-leftmost
         +::SKU slide-leftmost''', 'Date SKU Item Customer Quantity Unit Paid')



'VisiData: a curses interface for exploring and arranging tabular data'

__version__ = '3.3'
__version_info__ = 'VisiData v' + __version__
__author__ = 'Saul Pwanson <vd@saul.pw>'
__status__ = 'Production/Stable'
__copyright__ = 'Copyright (c) 2016-2024 ' + __author__


class EscapeException(BaseException):
    'Inherits from BaseException to avoid "except Exception" clauses.  Do not use a blanket "except:" or the task will be uncancelable.'
    pass


def addGlobals(*args, **kwargs):
    '''Update the VisiData globals dict with items from *args* and *kwargs*; to add symbols available to command execstrings and eval strings like command expr.'

    Dunder methods are ignored, to prevent accidentally overwriting housekeeping methods.'''
    drop_dunder = lambda d: {k: v for k, v in d.items() if not k.startswith("__")}
    for g in args:
        globals().update(drop_dunder(g))
    globals().update(drop_dunder(kwargs))


def getGlobals():
    'Return the VisiData globals dict.'
    return globals()

from .utils import *

from .extensible import *
from .vdobj import *
vd = VisiData()

vd.version = __version__

vd.addGlobals = addGlobals
vd.getGlobals = getGlobals

import visidata.keys

from .basesheet import *

import visidata.settings

# importModule tracks where commands/options/etc are coming from (via vd.importingModule)
core_imports = '''
import visidata.errors
import visidata.editor
import visidata.color
import visidata.cliptext
import visidata.mainloop

import visidata.menu
import visidata.wrappers
import visidata.undo
import visidata._types
import visidata.column

import visidata.interface
import visidata.sheets
import visidata.rename_col
import visidata.indexsheet

import visidata.statusbar

import visidata.textsheet
import visidata.threads
import visidata.path
import visidata.guide

import visidata.stored_list
import visidata._input
import visidata.tuiwin
import visidata.mouse
import visidata.movement

import visidata.type_date

import visidata._urlcache
import visidata.selection
import visidata.text_source
import visidata.loaders
import visidata.loaders.tsv
import visidata.pyobj
import visidata.loaders.json
import visidata._open
import visidata.save
import visidata.search

import visidata.expr
import visidata.metasheets
import visidata.input_history
import visidata.optionssheet
import visidata.type_currency
import visidata.type_floatsi
import visidata.clean_names
import visidata.cmdlog
import visidata.clipboard
import visidata.choose
import visidata.aggregators
import visidata.pivot
import visidata.freqtbl
import visidata.canvas
import visidata.canvas_text
import visidata.graph
import visidata.motd
import visidata.shell
import visidata.main
import visidata.help
import visidata.modify
import visidata.sort
import visidata.memory
import visidata.macros

import visidata.macos
import visidata.windows

import visidata.form
import visidata.sidebar

import visidata.ddwplay
import visidata.plugins

import visidata.theme
import visidata.apps
import visidata.fuzzymatch
import visidata.hint
'''

for line in core_imports.splitlines():
    if not line: continue
    module = line[len('import '):]
    vd.importModule(module)

vd.importSubmodules('visidata.loaders')

def importFeatures():
    vd.importSubmodules('visidata.features')
    vd.importSubmodules('visidata.themes')

    import visidata.deprecated

    vd.importModule('copy', 'copy deepcopy'.split())
    vd.importModule('builtins', 'abs all any ascii bin bool bytes callable chr complex dict dir divmod enumerate eval filter float format getattr hex int len list map max min next oct ord pow range repr reversed round set sorted str sum tuple type zip'.split())
    vd.importModule('math', 'acos acosh asin asinh atan atan2 atanh ceil copysign cos cosh degrees dist erf erfc exp expm1 fabs factorial floor fmod frexp fsum gamma gcd hypot isclose isfinite isinf isnan isqrt lcm ldexp lgamma log log1p log10 log2 modf radians remainder sin sinh sqrt tan tanh trunc prod perm comb nextafter ulp pi e tau inf nan'.split())
    vd.importModule('random', 'randrange randint choice choices sample uniform gauss lognormvariate'.split())
    vd.importModule('string', 'ascii_letters ascii_lowercase ascii_uppercase digits hexdigits punctuation printable whitespace'.split())
    vd.importModule('json')
    vd.importModule('itertools')
    vd.importModule('curses')

    import visidata.experimental  # import nothing by default but make package accessible

vd.finalInit()  # call all VisiData.init() from modules

importFeatures()

vd.addGlobals(vd=vd) # globals())



from visidata import vd, VisiData, Sheet, ItemColumn, asyncthread


vd._inputHistoryList = vd.StoredList(name='input_history')
vd.inputHistory = {}   # [input_type][input] -> anything


@VisiData.api
def addInputHistory(vd, input:str, type:str=''):
    r = vd.processInputHistory(input, type)
    if r:
        vd._inputHistoryList.append(r)


@VisiData.api
def processInputHistory(vd, input:str, type:str=''):
    hist = list(vd.inputHistory.setdefault(type, {}).keys())
    if hist and hist[-1] == input:
        return
    if input in vd.inputHistory[type]:
        n = vd.inputHistory[type][input].get('n', 0)
        del vd.inputHistory[type][input]  # make it the most recent entry
    else:
        n = 0

    r = dict(type=type, input=input, n=n+1)
    vd.inputHistory[type][input] = r
    return r


class InputHistorySheet(Sheet):
    # rowdef: dict(type=, input=, n=)
    # .source=vd.inputHistory
    columns = [
        ItemColumn('type'),
        ItemColumn('input'),
    ]
    def iterload(self):
        yield from vd._inputHistoryList


@VisiData.before
@asyncthread
def run(vd, *args, **kwargs):
    vd._inputHistoryList.reload()
    for x in vd._inputHistoryList:
        vd.processInputHistory(x.input, x.type)


@VisiData.property
def inputHistorySheet(vd):
    return InputHistorySheet('input_history', source=vd._inputHistoryList.path)


vd.addCommand(None, 'open-input-history', 'vd.push(inputHistorySheet)', 'open sheet with previous inputs')



from visidata import vd, BaseSheet

BaseSheet.addCommand('', 'repeat-last', 'execCommand(vd.cmdlog.rows[-1].longname) if vd.cmdlog.rows else fail("no recent command to repeat")', 'run most recent command with an empty, queried input')
BaseSheet.addCommand('', 'repeat-input', 'r = copy(vd.cmdlog.rows[-1]) if vd.cmdlog.rows else fail("no recent command to repeat"); vd.cmdlog.repeat_for_n(r, 1)', 'run previous modifying command (incl input)')
BaseSheet.addCommand('', 'repeat-input-n', 'r = copy(vd.cmdlog.rows[-1]) if vd.cmdlog.rows else fail("no recent command to repeat"); vd.cmdlog.repeat_for_n(r, input("# times to repeat prev command:", value=1))', 'run previous command (incl its input) N times')
BaseSheet.addCommand('', 'repeat-input-selected', 'r = copy(vd.cmdlog.rows[-1]) if vd.cmdlog.rows else fail("no recent command to repeat"); vd.cmdlog.repeat_for_selected(r)', 'run previous command (incl its input) for each selected row')

vd.addMenuItem('Edit', 'Repeat', 'last command', 'repeat-input')
vd.addMenuItem('Edit', 'Repeat', 'last command N times', 'repeat-input-n')
vd.addMenuItem('Edit', 'Repeat', 'last command for all selected rows', 'repeat-input-selected')



# command setdiff-sheet adds a diff colorizer for all sheets against current sheet

from visidata import Sheet, CellColorizer, vd

vd.theme_option('color_diff', 'red', 'color of values different from --diff source')
vd.theme_option('color_diff_add', 'yellow', 'color of rows/columns added to --diff source')


def makeDiffColorizer(othersheet):
    def colorizeDiffs(sheet, col, row, cellval):
        if row is None or col is None:
            return None
        vcolidx = sheet.visibleCols.index(col)
        rowidx = sheet.rows.index(row)
        if vcolidx < len(othersheet.visibleCols) and rowidx < len(othersheet.rows):
            otherval = othersheet.visibleCols[vcolidx].getDisplayValue(othersheet.rows[rowidx])
            if cellval.display != otherval:
                return 'color_diff'
        else:
            return 'color_diff_add'
    return colorizeDiffs


@Sheet.api
def setDiffSheet(vs):
    Sheet.colorizers.append(CellColorizer(8, None, makeDiffColorizer(vs)))


Sheet.addCommand(None, 'setdiff-sheet', 'setDiffSheet()', 'set this sheet as diff sheet for all new sheets')



from copy import copy
import itertools

from visidata import vd, vlen, VisiData, Column, AttrColumn, Sheet, ColumnsSheet, ENTER, Fanout
from visidata.pivot import PivotSheet, PivotGroupRow


vd.theme_option('disp_histogram', '■', 'histogram element character')
vd.option('histogram_bins', 0, 'number of bins for histogram of numeric columns')
vd.option('numeric_binning', False, 'bin numeric columns into ranges', replay=True)


@VisiData.api
def valueNames(vd, discrete_vals, numeric_vals):
    ret = [ '+'.join(str(x) for x in discrete_vals) ]
    if isinstance(numeric_vals, tuple) and numeric_vals != (0, 0):
        ret.append('%s-%s' % numeric_vals)

    return '+'.join(ret)

class HistogramColumn(Column):
    '.sourceCol is the column to be histogrammed'
    def calcValue(col, row):
        histogram = col.sheet.options.disp_histogram
        histolen = col.width-2
        return histogram*(histolen*col.sourceCol.getTypedValue(row)//col.largest)

    def updateLargest(col, row):
        col.largest = max(col.largest, col.sourceCol.getTypedValue(row))


def makeFreqTable(sheet, *groupByCols):
    if not any(groupByCols):
        vd.fail('FreqTableSheet requires at least 1 column for grouping')
    return FreqTableSheet(sheet.name,
                          '%s_freq' % '-'.join(col.name for col in groupByCols),
                          groupByCols=groupByCols,
                          source=sheet)


class FreqTableSheet(PivotSheet):
    'Generate frequency-table sheet on currently selected column.'
    guide = '''# Frequency Table
This is a *frequency analysis* of _{sheet.groupByColsName}_ from the *{sheet.groupByCols[0].sheet}* sheet.

Each row on this sheet corresponds to a *bin* of rows on the source sheet that have a distinct value.  The _count_ and _percent_ columns show how many rows on the source sheet are in this bin.

- `Enter` to open a copy of the source sheet, with only the rows in the current bin.
- `g Enter` to open a copy of the source sheet, with a combination of the rows from all selected bins.

## Tips

- Use `+` on the source sheet, to add aggregators on other columns, and those metrics will appear as separate columns here.
- Selecting bins on this sheet will select those rows on the source sheet.
'''
    rowtype = 'bins'  # rowdef FreqRow(keys, sourcerows)

    @property
    def groupByColsName(self):
        return '+'.join(c.name for c in self.groupByCols)

    def selectRow(self, row):
        # Does not create an undo-operation for the select on the source rows. The caller should create undo-information itself.
        self.source.select(row.sourcerows, status=False, add_undo=False)     # select all entries in the bin on the source sheet
        return super().selectRow(row)  # then select the bin itself on this sheet

    def unselectRow(self, row):
        self.source.unselect(row.sourcerows, status=False, add_undo=False)
        return super().unselectRow(row)

    def addUndoSelection(self):
        self.source.addUndoSelection()
        super().addUndoSelection()

    # override Sheet operations that handle multiple rows:
    #     select(), unselect(), and toggle()
    # to make undo more efficient. Without this optimization, the memory
    # use for the undo-tracking on the source sheet is O(n^2) in the number
    # of bins selected, which can easily exceed all available memory.
    def select(self, rows, status=True, progress=True, add_undo=True):
        if add_undo:
            self.addUndoSelection()
        super().select(rows, status, progress, add_undo=False)

    def unselect(self, rows, status=True, progress=True, add_undo=True):
        if add_undo:
            self.addUndoSelection()
        super().unselect(rows, status, progress, add_undo=False)

    def toggle(self, rows, add_undo=True):
        'Toggle selection of given *rows* and corresponding rows in source sheet.'
        if add_undo:
            self.addUndoSelection()
        super().toggle(rows, add_undo=False)

    def select_row(self, row, add_undo=True):
        'Add single *row* to set of selected rows, and corresponding rows in source sheet.'
        if add_undo:
            self.addUndoSelection()
        super().select_row(row, add_undo=False)

    def unselect_row(self, row, add_undo=True):
        'Remove single *row* from set of selected rows, and remove corresponding rows in source sheet.'
        if add_undo:
            self.addUndoSelection()
        super().unselect_row(row, add_undo=False)

    def toggle_row(self, row, add_undo=True):
        'Toggle selection of given *row* and of corresponding rows in source sheet.'
        if add_undo:
            self.addUndoSelection()
        super().toggle_row(row, add_undo=False)

    def resetCols(self):
        super().resetCols()

        # add default bonus columns
        countCol = AttrColumn('count', 'sourcerows', type=vlen)
        for c in [
            countCol,
            Column('percent', type=float, getter=lambda col,row: len(row.sourcerows)*100/col.sheet.source.nRows),
        ]:
            self.addColumn(c)

        if self.options.disp_histogram:
            c = HistogramColumn('histogram', type=str, width=self.options.default_width*2, sourceCol=countCol)
            self.addColumn(c)

        # if non-numeric grouping, reverse sort by count at end of load
        if not any(vd.isNumeric(c) for c in self.groupByCols):
            self._ordering = [(countCol, True)]

    def loader(self):
        'Generate frequency table.'
        # two more threads
        histcols = [col for col in self.visibleCols if isinstance(col, HistogramColumn)]
        vd.sync(self.addAggregateCols(),
                self.groupRows(lambda row, cols=Fanout(histcols): cols.updateLargest(row)))

    def afterLoad(self):
        super().afterLoad()
        if self.nCols > len(self.groupByCols)+3:  # hide percent/histogram if aggregations added
            self.column('percent').hide()
            self.column('histogram').hide()

    def openRow(self, row):
        'open copy of source sheet with rows that are grouped in current row'
        if row.sourcerows:
            vs = copy(self.source)
            vs.names = vs.names + [vd.valueNames(row.discrete_keys, row.numeric_key)]
            vs.rows=copy(row.sourcerows)
            return vs
        vd.warning("no source rows")

    def openRows(self, rows):
        vs = copy(self.source)
        vs.names = vs.names + ["several"]
        vs.source = self
        vs.rows = list(itertools.chain.from_iterable(row.sourcerows for row in rows))
        return vs

    def openCell(self, col, row):
        return Sheet.openCell(self, col, row)


class FreqTableSheetSummary(FreqTableSheet):
    'Append a PivotGroupRow to FreqTableSheet with only selectedRows.'
    def afterLoad(self):
        self.addRow(PivotGroupRow(['Selected'], (0,0), self.source.selectedRows, {}))
        super().afterLoad()


def makeFreqTableSheetSummary(sheet, *groupByCols):
    return FreqTableSheetSummary(sheet.name,
                          '%s_freq' % '-'.join(col.name for col in groupByCols),
                          groupByCols=groupByCols,
                          source=sheet)


@VisiData.api
class FreqTablePreviewSheet(Sheet):
    @property
    def rows(self):
        return self.source.cursorRow.sourcerows


FreqTableSheet.addCommand('', 'open-preview', 'vd.push(FreqTablePreviewSheet(sheet.name, "preview", source=sheet, columns=source.columns), pane=2); vd.options.disp_splitwin_pct=50', 'open split preview of source rows at cursor')

Sheet.addCommand('F', 'freq-col', 'vd.push(makeFreqTable(sheet, cursorCol))', 'open Frequency Table grouped on current column, with aggregations of other columns')
Sheet.addCommand('gF', 'freq-keys', 'vd.push(makeFreqTable(sheet, *keyCols)) if keyCols else vd.fail("there are no key columns to group by")', 'open Frequency Table grouped by all key columns on source sheet, with aggregations of other columns')
Sheet.addCommand('zF', 'freq-summary', 'vd.push(makeFreqTableSheetSummary(sheet, Column("Total", sheet=sheet, getter=lambda col, row: "Total")))', 'open one-line summary for all rows and selected rows')

ColumnsSheet.addCommand(ENTER, 'freq-row', 'vd.push(makeFreqTable(source[0], cursorRow))', 'open a Frequency Table sheet grouped on column referenced in current row')
vd.addMenuItem('Data', 'Frequency table', 'current row', 'freq-row')

FreqTableSheet.addCommand('gu', 'unselect-rows', 'unselect(selectedRows)', 'unselect all source rows grouped in current row')
FreqTableSheet.addCommand('g'+ENTER, 'dive-selected', 'vd.push(openRows(selectedRows))', 'open copy of source sheet with rows that are grouped in selected rows')
FreqTableSheet.addCommand('', 'select-first', 'for r in rows: source.select([r.sourcerows[0]])', 'select first source row in each bin')

HistogramColumn.init('largest', lambda: 1)

vd.addGlobals(
    makeFreqTable=makeFreqTable,
    makeFreqTableSheetSummary=makeFreqTableSheetSummary,
    FreqTableSheet=FreqTableSheet,
    FreqTableSheetSummary=FreqTableSheetSummary,
    HistogramColumn=HistogramColumn,
)

vd.addMenuItems('''
    Data > Frequency table > current column > freq-col
    Data > Frequency table > key columns > freq-keys
''')



from visidata import vd, VisiData, BaseSheet, Sheet, Column, AttrColumn, CellColorizer, Option


@BaseSheet.lazy_property
def optionsSheet(sheet):
    return OptionsSheet(sheet.name+"_options", source=sheet)

@VisiData.lazy_property
def globalOptionsSheet(vd):
    return OptionsSheet('global_options', source='global')


@VisiData.api
class OptionsSheet(Sheet):
    _rowtype = Option  # rowdef: Option
    rowtype = 'options'
    precious = False
    columns = (
        Column('option', getter=lambda col,row: row.name),
        Column('module', getter=lambda col,row: row.module, disp_expert=1),
        Column('value',
            getter=lambda col,row: col.sheet.diffOption(row.name),
            setter=lambda col,row,val: vd.options.set(row.name, val, col.sheet.source)
            ),
        Column('default', getter=lambda col,row: vd.options.getdefault(row.name)),
        Column('description', width=40, getter=lambda col,row: vd.options._get(row.name, 'default').helpstr),
        AttrColumn('replayable', disp_expert=1),
    )
    colorizers = [
        CellColorizer(3, None, lambda s,c,r,v: v.value if r and c in s.columns[2:4] and r.name.startswith('color_') else None),
    ]
    nKeys = 2

    @property
    def guide(self):
        if self.source == 'global':
            r = '# Global Options\nThis is a list of global option settings.'
        else:
            r = '# Sheet Options\nThis is a list of option settings specifically for the current sheet.'

        r += f'\n\n- `e` to edit/toggle the current option value'
        r += '\n- `d` to restore option to builtin default'
        return r

    def diffOption(self, optname):
        return vd.options.getonly(optname, self.source, '')

    def editOption(self, row):
        currentValue = vd.options.getobj(row.name, self.source)
        vd.addUndo(vd.options.set, row.name, currentValue, self.source)
        if isinstance(row.value, bool):
            vd.options.set(row.name, not currentValue, self.source)
        else:
            helpstr = f'# options.{self.cursorRow.name}'
            opt = vd.options._get(self.cursorRow.name, 'default')
            if opt.helpstr:
                x = getattr(vd, 'help_'+opt.helpstr, opt.helpstr or '')
                helpstr += (x or '').strip()
            if opt.extrahelp:
                helpstr += '\n'+opt.extrahelp.strip()
            valcolidx = self.visibleCols.index(self.column(self.valueColName))
            v = self.editCell(valcolidx, value=currentValue, help=helpstr)
            vd.options.set(row.name, v, self.source)

    @property
    def valueColName(self):
        return 'global_value' if self.source == 'global' else 'sheet_value'

    def beforeLoad(self):
        super().beforeLoad()
        self.columns[2].name = self.valueColName

    def iterload(self):
        for k in vd.options.keys():
            v = vd.options._get(k)
            if v.sheettype in [None, BaseSheet]:
                yield v
            elif self.source != 'global' and v.sheettype in self.source.superclasses():
                yield v

    def newRow(self):
        vd.fail('adding rows to the options sheet is not supported.')


BaseSheet.addCommand('O', 'options-global', 'vd.push(vd.globalOptionsSheet)', 'open Options Sheet: edit global options (apply to all sheets)')

BaseSheet.addCommand('zO', 'options-sheet', 'vd.push(sheet.optionsSheet)', 'open Options Sheet: edit sheet options (apply to current sheet only)')

OptionsSheet.addCommand('d', 'unset-option', 'options.unset(cursorRow.name, str(source))', 'remove option override for this context')
OptionsSheet.addCommand(None, 'edit-option', 'editOption(cursorRow)', 'edit option at current row')
OptionsSheet.bindkey('e', 'edit-option')
OptionsSheet.bindkey('Enter', 'edit-option')

vd.addMenuItems('''
    File > Options > all sheets > options-global
    File > Options > this sheet > options-sheet
''')



import os
import os.path
import sys

from visidata import VisiData, vd, Path, BaseSheet, TableSheet, TextSheet, SettableColumn


vd.option('filetype', '', 'specify file type', replay=True)


@VisiData.api
def inputFilename(vd, prompt, *args, **kwargs):
    completer= _completeFilename
    if not vd.couldOverwrite():  #1805 don't suggest an existing file
        completer = None
        v = kwargs.get('value', '')
        if v and Path(v).exists():
            kwargs['value'] = ''
    return vd.input(prompt, type="filename", *args, completer=completer, **kwargs).strip()


@VisiData.api
def inputPath(vd, *args, **kwargs):
    return Path(vd.inputFilename(*args, **kwargs))


def _completeFilename(val, state):
    i = val.rfind('/')
    if i < 0:  # no /
        base = ''
        partial = val
    elif i == 0: # root /
        base = '/'
        partial = val[1:]
    else:
        base = val[:i]
        partial = val[i+1:]

    files = []
    for f in os.listdir(Path(base or '.')):
        if f.startswith(partial):
            files.append(os.path.join(base, f))

    files.sort()
    return files[state%len(files)]


@VisiData.api
def guessFiletype(vd, p, *args, funcprefix='guess_'):
    '''Call all vd.guess_<filetype>(p) functions and return best candidate sheet based on file contents.'''

    guessfuncs = [getattr(vd, x) for x in dir(vd) if x.startswith(funcprefix)]
    filetypes = []
    for f in guessfuncs:
        try:
            filetype = f(p, *args)
            if filetype:
                filetype['_guesser'] = f.__name__
                filetypes.append(filetype)
        except FileNotFoundError:
            pass
        except Exception as e:
            vd.debug(f'{f.__name__}: {e}')

    if filetypes:
        return sorted(filetypes, key=lambda r: -r.get('_likelihood', 1))[0]

    return {}


@VisiData.api
def guess_extension(vd, path):
    # try auto-detect from extension
    ext = path.suffix[1:].lower()
    openfunc = getattr(vd, f'open_{ext}', vd.getGlobals().get(f'open_{ext}'))
    if openfunc:
        return dict(filetype=ext, _likelihood=3)


@VisiData.api
def openPath(vd, p, filetype=None, create=False):
    '''Call ``open_<filetype>(p)`` or ``openurl_<p.scheme>(p, filetype)``.  Return constructed but unloaded sheet of appropriate type.
    If True, *create* will return a new, blank **Sheet** if file does not exist.'''
    # allow user to assign a filetype to a pathname:  options.set('filetype', 'csv', '-')
    filetype = filetype or vd.options.getonly('filetype', str(p), None)  #1710
    filetype = filetype or vd.options.getonly('filetype', 'global', None)

    if p.scheme and not p.has_fp():
        schemes = p.scheme.split('+')
        openfuncname = 'openurl_' + schemes[-1]

        openfunc = getattr(vd, openfuncname, None) or vd.getGlobals().get(openfuncname, None)
        if not openfunc:
            vd.fail(f'no loader for url scheme: {p.scheme}')

        return openfunc(p, filetype=filetype)

    if not p.exists() and not create:
        return None

    # assign filetype from extension, but only for files, not directories
    if not p.is_dir():  #2547
        filetype = filetype or p.ext
    filetype = filetype or vd.options.filetype

    filetype = filetype.lower()

    if not p.exists():
        newfunc = getattr(vd, 'new_' + filetype, vd.getGlobals().get('new_' + filetype))
        if not newfunc:
            vd.warning('%s does not exist, creating new sheet' % p)
            return vd.newSheet(p.base_stem, 1, source=p)

        vd.status('creating blank %s' % (p.given))
        return newfunc(p)

    if p.is_fifo():
        # read the file as text, into a RepeatFile that can be opened multiple times
        p = Path(p.given, fp=p.open(mode='rb'))

    openfuncname = 'open_' + filetype
    openfunc = getattr(vd, openfuncname, vd.getGlobals().get(openfuncname))
    if not openfunc:
        opts = vd.guessFiletype(p)
        if opts and 'filetype' in opts:
            filetype = opts['filetype']
            openfuncname = 'open_' + filetype
            openfunc = getattr(vd, openfuncname, vd.getGlobals().get(openfuncname))
            if not openfunc:
                vd.error(f'guessed {filetype} but no {openfuncname}')

            vs = openfunc(p)
            for k, v in opts.items():
                if k != 'filetype' and not k.startswith('_'):
                    setattr(vs.options, k, v)
            vd.warning('guessed "%s" filetype based on contents' % opts['filetype'])
            return vs

        vd.warning('unknown "%s" filetype' % filetype)

        filetype = 'txt'
        openfunc = vd.open_txt

    vd.status('opening %s as %s' % (p.given, filetype))

    return openfunc(p)

@VisiData.api
def openSource(vd, p, filetype=None, create=False, **kwargs):
    '''Return unloaded sheet object for *p* opened as the given *filetype* and with *kwargs* as option overrides. *p* can be a Path or a string (filename, url, or "-" for stdin).
    when true, *create* will return a blank sheet, if file does not exist.'''

    if isinstance(p, BaseSheet):
        return p

    vs = None
    if isinstance(p, str):
        if '://' in p:
            vs = vd.openPath(Path(p), filetype=filetype)  # convert to Path and recurse
        elif p == '-':
            if vd.stdinSource.fptext.isatty():
                vd.fail('cannot open stdin when it is a tty')
            vs = vd.openPath(vd.stdinSource, filetype=filetype)
        else:
            vs = vd.openPath(Path(p), filetype=filetype, create=create)  # convert to Path and recurse
    else:
        vs = vd.openPath(p, filetype=filetype, create=create)

    for optname, optval in kwargs.items():
        vs.options[optname] = optval

    return vs


#### enable external addons
@VisiData.api
def open_txt(vd, p):
    'Create sheet from `.txt` file at Path `p`, checking whether it is TSV.'
    if p.exists(): #1611
        with p.open(encoding=vd.options.encoding) as fp:
            delimiter = vd.options.delimiter
            try:
                if delimiter and delimiter in next(fp):    # peek at the first line
                    return vd.open_tsv(p)  # TSV often have .txt extension
            except StopIteration:
                return vd.newSheet(p.base_stem, 1, source=p)
    return TextSheet(p.base_stem, source=p)


BaseSheet.addCommand('o', 'open-file', 'vd.push(openSource(inputFilename("open: "), create=True))', 'Open file or URL')
TableSheet.addCommand('zo', 'open-cell-file', 'vd.push(openSource(cursorDisplay) or fail(f"file {cursorDisplay} does not exist"))', 'Open file or URL from path in current cell')
BaseSheet.addCommand('gU', 'undo-last-quit', 'push(allSheets[-1])', 'reopen most recently closed sheet')

vd.addMenuItems('''
    File > Open > input file/url > open-file
    File > Reopen last closed > undo-last-quit
''')



import collections
import functools
import sys
import inspect
import argparse
import importlib
import os

import visidata
from visidata import VisiData, BaseSheet, vd, AttrDict
from visidata.vendor.appdirs import user_config_dir, user_cache_dir


# [settingname] -> { objname(Sheet-instance/Sheet-type/'global'/'default'): Option/Command/longname }
class SettingsMgr(collections.OrderedDict):
    def __init__(self):
        super().__init__()
        self.allobjs = {}

    def __hash__(self):
        return hash(id(self))

    def __eq__(self, other):
        return self is other

    def objname(self, obj):
        if isinstance(obj, str):
            v = obj
        elif obj is None:
            v = 'global'
        elif isinstance(obj, BaseSheet):
            v = obj.name
        elif inspect.isclass(obj) and issubclass(obj, BaseSheet):
            v = obj.__name__
        else:
            return None

        self.allobjs[v] = obj
        return v

    def getobj(self, objname):
        'Inverse of objname(obj); returns obj if available'
        return self.allobjs.get(objname)

    def unset(self, k, obj='default'):
        'Remove setting for given key in the given context.'
        objstr = self.objname(obj)
        if objstr in self[k]:
            del self[k][objstr]

    def set(self, k, v, obj):
        'obj is a Sheet instance, or a Sheet [sub]class.  obj="global" means override default unless there is a sheet-specific override; obj="default" means last resort.'
        if k not in self:
            self[k] = dict()
        self[k][self.objname(obj)] = v
        return v

    def setdefault(self, k, v):
        return self.set(k, v, 'default')

    @functools.lru_cache()
    def _mappings(self, obj):
        '''Return list of contexts in order to resolve settings. ordering is, from lowest to highest precedence:

        1. "default": default specified in option() definition
        2. "global": in order of program execution:
            a. .visidatarc
            b. command-line options, applied on top of the overrides in .visidatarc
            c. at runtime via 'O'ptions meta-sheet
        3. objname(type(obj)): current sheet class and parents, recursively
        4. objname(obj): the specific sheet instance
            a. can override at runtime, replace value for sheet instance
        '''
        mappings = []
        if obj:
            mappings += [obj]
            mappings += [self.objname(cls) for cls in inspect.getmro(type(obj))]

        mappings += ['global', 'default']
        return mappings

    def _get(self, key, obj=None):
        d = self.get(key, None)
        if d:
            for m in self._mappings(obj or vd.activeSheet):
                v = d.get(self.objname(m))
                if v is not None:
                    return v

    def iter(self, obj=None):
        'Iterate through all keys considering context of obj. If obj is None, uses the context of the top sheet.'
        if obj is None and vd:
            obj = vd.activeSheet

        for o in self._mappings(obj):
            o = self.objname(o)
            for k in self.keys():
                for o2 in self[k]:
                    if o == o2:
                        yield (k, o), self[k][o2]

    def iterall(self):
        for k in self.keys():
            for o in self[k]:
                yield (k, o), self[k][o]



class Command:
    def __init__(self, longname, execstr, helpstr='', module='', replay=True, deprecated=False):
        self.longname = longname
        self.execstr = execstr
        self.helpstr = helpstr
        self.module = module
        self.deprecated = deprecated
        self.replayable = replay


class Option:
    def __init__(self, name, value, description='', module='', help=''):
        # description gets shows on the manpage and the optionssheet; help is shown on the sidebar while editing
        self.name = name
        self.value = value
        self.helpstr = description
        self.extrahelp = help
        self.replayable = False
        self.sheettype = BaseSheet
        self.module = module

    def __str__(self):
        return str(self.value)

    def __eq__(self, other):
        return self.name == other.name


@VisiData.api
class OptionsObject:
    'minimalist options framework'
    def __init__(self, mgr, obj=None):
        object.__setattr__(self, '_opts', mgr)
        object.__setattr__(self, '_cache', {})
        object.__setattr__(self, '_obj', obj)

    def keys(self, obj=None):
        for k, d in self._opts.items():
            if obj is None or self._opts.objname(obj) in d:
                yield k

    def _get(self, k, obj=None):
        'Return Option object for k in context of obj. Cache result until any set().'
        opt = self._cache.get((k, obj or vd.activeSheet), None)
        if opt is None:
            opt = self._opts._get(k, obj)
            self._cache[(k, obj or vd.activeSheet)] = opt
        return opt

    def _set(self, k, v, obj=None, helpstr='', module=None):
        k, v = vd._resolve_optalias(k, v)  # to set deprecated and abbreviated options

        opt = self._get(k) or Option(k, v, '', module)
        self._cache.clear()  # invalidate entire cache on any change
        return self._opts.set(k, Option(k, v, opt.helpstr or helpstr, opt.module or module), obj)

    def is_set(self, k, obj=None):
        d = self._opts.get(k, None)
        if d:
            return d.get(self._opts.objname(obj), None)

    def get(self, optname, default=None):
        'Return the value of the given *optname* option in the options context. *default* is only returned if the option is not defined.  An Exception is never raised.'
        d = self._get(optname, None)
        if d:
            return d.value
        return default

    def getobj(self, optname, obj=None):
        'Return value of option optname as set on obj, or on option context if obj is None.'
        return self._get(optname, obj).value

    def getdefault(self, optname):
        return self._get(optname, 'default').value

    def getonly(self, optname, obj, default):
        'Return value of option optname as set on obj, or default if not set specifically on obj'
        d = self._opts.get(optname, None)
        if d:
            opt = d.get(self._opts.objname(obj), None)
            if opt:
                return opt.value
        return default

    def set(self, optname, value, obj='global'):
        "Override *value* for *optname* in the options context, or in the *obj* context if given."
        opt = self._get(optname)
        module = None  # keep default
        if opt:
            curval = opt.value
            t = type(curval)
            if value is None and curval is not None:
                return self.unset(optname, obj=obj)
            elif isinstance(value, str) and t is bool: # special case for bool options
                value = value and (value[0] not in "0fFnN")  # ''/0/false/no are false, everything else is true
            elif type(value) is t:    # if right type, no conversion
                pass
            elif curval is None:  # if None, do not apply type conversion
                pass
            else:
                value = t(value)

            if curval != value and self._get(optname, 'default').replayable:
                if obj != 'default' and type(obj) is not type:  # default and class options set on init aren't recorded
                    if vd.cmdlog:
                        self.add_option_to_cmdlogs(obj, optname, value, 'set-option')
        else:
            curval = None
            vd.warning('setting unknown option %s' % optname)
            module = 'unknown'

        return self._set(optname, value, obj, module=module)

    def unset(self, optname, obj=None):
        'Remove setting value for given context.'
        v = self._opts.unset(optname, obj)
        opt = self._get(optname)
        if vd.cmdlog and opt and opt.replayable:
            self.add_option_to_cmdlogs(obj, optname, value='', longname='unset-option')
        self._cache.clear()  # invalidate entire cache on any change
        return v

    def add_option_to_cmdlogs(self, obj, optname, value='', longname='set-option'):
        'Records option-set on cmdlogs'
        objname = self._opts.objname(obj)
        # all options are recorded on global cmdlog
        vd.cmdlog.addRow(vd.cmdlog.newRow(sheet=objname, row=optname,
                    keystrokes='', input=str(value),
                    longname=longname, undofuncs=[]))
        # global options are recorded on all cmdlog_sheet's
        if obj == 'global':
            for vs in vd.sheets:
                vs.cmdlog_sheet.addRow(vd.cmdlog.newRow(sheet=objname, row=optname,
                            keystrokes='', input=str(value),
                            longname=longname, undofuncs=[]))
        # sheet-specific options are recorded on that sheet
        elif isinstance(obj, BaseSheet):
            obj.cmdlog_sheet.addRow(vd.cmdlog.newRow(sheet=objname, row=optname,
                        keystrokes='', input=str(value),
                        longname=longname, undofuncs=[]))

    def setdefault(self, optname, value, helpstr, module):
        return self._set(optname, value, 'default', helpstr=helpstr, module=module)

    def getall(self, prefix=''):
        'Return dictionary of all options beginning with `prefix` (with `prefix` removed from the name).'
        return { optname[len(prefix):] : vd.options[optname]
                    for optname in vd.options.keys()
                        if optname.startswith(prefix) }

    def __getattr__(self, optname):      # options.foo
        'Return value of option `optname` for stored options context.'
        return self.__getitem__(optname)

    def __setattr__(self, optname, value):   # options.foo = value
        'Set *value* of option *optname* for stored options context.'
        self.__setitem__(optname, value)

    def __getitem__(self, optname):      # options[optname]
        opt = self._get(optname, obj=self._obj)
        if not opt:
            raise ValueError('no option "%s"' % optname)
        return opt.value

    def __setitem__(self, optname, value):   # options[optname] = value
        self.set(optname, value, obj=self._obj)


vd.commands = SettingsMgr()
vd.bindkeys = SettingsMgr()
vd._options = SettingsMgr()

vd.options = vd.OptionsObject(vd._options)  # global option settings
vd.option_aliases = {}


@VisiData.api
def optalias(vd, altname, optname, val=None):
    'Create an alias `altname` for option `optname`, setting the value to a particular `val` (if not None).'
    vd.option_aliases[altname] = (optname, val)


@VisiData.api
def _resolve_optalias(vd, optname, optval):
    while optname in vd.option_aliases:
        optname, v = vd.option_aliases.get(optname)
        if v is not None:  # value might be given
            optval = v

    return optname, optval


@VisiData.api
def option(vd, name, default, description, replay=False, sheettype=BaseSheet, help:str=''):
    '''Declare a new option.

   - `name`: name of option
   - `default`: default value when no other override exists
   - `helpstr`: short description of option (as shown in the **Options Sheet**)
   - `replay`: ``True`` if changes to the option should be stored in the **Command Log**
   - `sheettype`: ``None`` if the option is not sheet-specific, to make it global on CLI
    '''
    opt = vd.options.setdefault(name, default, description, vd.importingModule)
    opt.replayable = replay
    opt.sheettype=sheettype
    opt.extrahelp = help
    return opt


@VisiData.api
def theme_option(vd, name, *args, **kwargs):
    if name.startswith('color_'):
        kwargs.setdefault('help', vd.help_color)
    return vd.option(name, *args, **kwargs)


@BaseSheet.class_api
@classmethod
def addCommand(cls, keystrokes, longname, execstr, helpstr='', replay=True, **kwargs):
    '''Add a new command to *cls* sheet type.

    - *keystrokes*: default keybinding, including **prefixes**.
    - *longname*: name of the command.
    - *execstr*: Python statement to pass to `exec()`'ed when the command is executed.
    - *helpstr*: help string shown in the **Commands Sheet**.
    '''
    cmd = Command(longname, execstr, helpstr=helpstr, module=vd.importingModule, replay=replay, **kwargs)
    vd.commands.set(longname, cmd, cls)
    if keystrokes:
        vd.bindkey(keystrokes, longname, cls)
    return longname

def _command(cls, binding, longname, helpstr, **kwargs):
    def decorator(func):
        funcname = longname.replace('-', '_')
        setattr(vd, funcname, func)
        cls.addCommand(binding, longname, f'vd.{funcname}(sheet)', helpstr, **kwargs)
    return decorator

BaseSheet.command = classmethod(_command)
globalCommand = BaseSheet.addCommand   # to be deprecated


@VisiData.api
def bindkey(vd, keystrokes, longname, obj='BaseSheet'):
    'Bind *keystrokes* to *longname* on BaseSheet and unbind more-specific bindings of keystrokes.'
    vd.bindkeys.set(vd.prettykeys(keystrokes.replace(' ', '')), longname, obj)

@VisiData.api
def unbindkey(vd, keystrokes, obj='BaseSheet'):
    'Bind *keystrokes* to *longname* on BaseSheet and unbind more-specific bindings of keystrokes.'
    vd.bindkeys.unset(vd.prettykeys(keystrokes.replace(' ', '')), obj)


@BaseSheet.class_api
@classmethod
def bindkey(cls, keystrokes, longname):
    'Bind *keystrokes* to *longname* on the *cls* sheet type.'
    oldlongname = vd.bindkeys._get(keystrokes, cls)
    if oldlongname:
        vd.warning('%s was already bound to %s' % (keystrokes, oldlongname))
    vd.bindkey(keystrokes, longname, cls)

@BaseSheet.class_api
@classmethod
def unbindkey(cls, keystrokes):
    '''Unbind `keystrokes` on a `<SheetType>`.
    May be necessary to avoid a warning when overriding a binding on the same exact class.'''
    vd.unbindkey(keystrokes, cls)

@BaseSheet.api
def getCommand(sheet, cmd):
    'Return the Command for the given *cmd*, which may be keystrokes, longname, or a Command itself, within the context of `sheet`.'
    if isinstance(cmd, Command):
        return cmd

    longname = cmd
    while vd.bindkeys._get(longname, obj=sheet) is not None:
        longname = vd.bindkeys._get(longname, obj=sheet)

    return vd.commands._get(longname, obj=sheet)


@VisiData.api
def loadConfigFile(vd, fn=''):
    p = visidata.Path(fn or vd.options.config)
    if p.exists():
        newdefs = {}
        try:
            with open(p) as fd:
                code = compile(fd.read(), str(p), 'exec')
            vd.importingModule = 'visidatarc'
            exec(code, vd.getGlobals(), newdefs)
        except Exception as e:
            vd.exceptionCaught(e)
        finally:
            vd.importingModule = None

        keys = newdefs.get('__all__', None)
        if keys:
            vd.addGlobals({k:newdefs[k] for k in keys})
        else:
            vd.addGlobals(newdefs)


def addOptions(parser):
    for optname in vd.options.keys('default'):
        if optname.startswith('color_') or optname.startswith('disp_'):
            continue
        action = 'store_true' if options[optname] is False else 'store'
        try:
            parser.add_argument('--' + optname.replace('_', '-'), action=action, dest=optname, default=None, help=options._opts._get(optname).helpstr)
        except argparse.ArgumentError:
            pass


def _get_config_file():
    xdg_config_file = visidata.Path(user_config_dir('visidata')) / 'config.py'
    if xdg_config_file.exists():
        return xdg_config_file
    else:
        return visidata.Path('~/.visidatarc')


def _get_cache_dir():
    return visidata.Path(user_cache_dir('visidata'))


@VisiData.api
def loadConfigAndPlugins(vd, args=AttrDict()):
    # set visidata_dir and config manually before loading config file, so visidata_dir can be set from cli or from $VD_DIR
    vd.options.visidata_dir = args.visidata_dir if args.visidata_dir is not None else os.getenv('VD_DIR', '') or vd.options.visidata_dir
    vd.options.config = args.config if args.config is not None else os.getenv('VD_CONFIG', '') or vd.options.config

    sys.path.append(str(visidata.Path(vd.options.visidata_dir)))
    sys.path.append(str(visidata.Path(vd.options.visidata_dir)/"plugins-deps"))

    # autoload installed plugins first
    args_plugins_autoload = args.plugins_autoload if 'plugins_autoload' in args else True
    if not args.nothing and args_plugins_autoload and vd.options.plugins_autoload:
        from importlib.metadata import entry_points
        eps_visidata = []
        try:
            eps = entry_points()
            vp = 'visidata.plugins'
            if hasattr(eps, 'groups'): #Python >= 3.10
                eps_visidata = eps.select(group=vp)
            else:                      #Python <  3.10
                eps_visidata = eps.get(vp, [])
        except Exception as e:
            vd.warning(f'plugin autoload failed; see issue #1529:  {e}')

        for ep in eps_visidata:
            try:
                vd.importingModule = ep.name
                plug = ep.load()
                sys.modules[f'visidata.plugins.{ep.name}'] = plug
                vd.debug(f'Plugin {ep.name} loaded')
            except Exception as e:
                vd.warning(f'Plugin {ep.name} failed to load')
                vd.exceptionCaught(e)
            finally:
                vd.importingModule = None

        # import plugins from .visidata/plugins before .visidatarc, so plugin options can be overridden
        for modname in (args.imports or vd.options.imports or '').split():
            try:
                vd.addGlobals(importlib.import_module(modname).__dict__)
            except ModuleNotFoundError as e:  #1131
                if 'plugins' in e.args[0]:
                    continue
                vd.exceptionCaught(e)
            except Exception as e:
                vd.exceptionCaught(e)
                continue

    # user customisations in config file in standard location
    if vd.options.config:
        vd.loadConfigFile(vd.options.config)


@VisiData.api
def importModule(vd, pkgname, symbols=[]):
    'Import the given *pkgname*, setting vd.importingModule to *pkgname* before import and resetting to None after.'
    modparts = pkgname.split('.')
    vd.importingModule = modparts[-1]
    r = importlib.import_module(pkgname)
    vd.importingModule = None
    vd.importedModules.append(r)
    vd.addGlobals({pkgname:r})
    if symbols:
        vd.addGlobals({k:getattr(r, k) for k in symbols if hasattr(r, k)})

    return r


@VisiData.api
def importSubmodules(vd, pkgname):
    'Import all files below the given *pkgname*'
    import pkgutil

    m = vd.importModule(pkgname)
    for module in pkgutil.walk_packages(m.__path__):
        vd.importModule(pkgname + '.' + module.name)


@VisiData.api
def importExternal(vd, modname, pipmodname=''):
    pipmodname = pipmodname or modname
    try:
        m = importlib.import_module(modname)
        vd.addGlobals({modname:m})
        return m
    except ModuleNotFoundError as e:
        vd.fail(f'External package "{modname}" not installed; run: pip install {pipmodname}')


@VisiData.api
def requireOptions(vd, *args, help=''):
    '''Prompt user to input values for option names in *args* if current values
    are non-false.  Offer to persist the values to visidatarc.'''

    optvals = {}
    for optname in args:
        if not getattr(vd.options, optname):
            if help:
                vd.status(help)
            v = vd.input(f'{optname}: ', record=False, display='password' not in optname)
            optvals[optname] = v

    vd.setPersistentOptions(**optvals)


@VisiData.api
def setPersistentOptions(vd, **kwargs):
    '''Set options from *kwargs* and offer to save them to visidatarc.'''
    for optname, optval in kwargs.items():
        setattr(vd.options, optname, optval)

    optnames = ' '.join(kwargs.keys())
    yn = vd.input(f'Save {len(kwargs)} options ({optnames}) to {vd.options.config}? ', record=False)[0:1]

    if yn and yn in 'Yy':
        with open(str(visidata.Path(vd.options.config)), mode='a') as fp:
            for optname, optval in kwargs.items():
                fp.write(f'options.{optname}={repr(optval)}\n')


vd.option('visidata_dir', user_config_dir('visidata'), 'directory to load and store additional files', sheettype=None)

BaseSheet.bindkey('^M', '^J')  # for windows ENTER

vd.addGlobals({
    'options': vd.options,  # legacy
    'globalCommand': BaseSheet.addCommand,
    'Option': Option,
})



'''
Add svg saver to Canvas.

Requires matplotlib.pyplot
'''


import collections

from visidata import VisiData, Canvas, vd, Progress

vd.option('plt_marker', '.', 'matplotlib.markers')


@Canvas.api
def plot_sheet(self, ax):
    plt = vd.importExternal('matplotlib.pyplot', 'matplotlib')
    nerrors = 0
    nplotted = 0

    self.reset()

    vd.status('loading data points')
    catcols = [c for c in self.xcols if not vd.isNumeric(c)]
    numcols = [c for c in self.xcols if vd.isNumeric(c)]
    for ycol in self.ycols:
        xpts = collections.defaultdict(list)
        ypts = collections.defaultdict(list)
        for rownum, row in enumerate(Progress(self.sourceRows, 'plotting')):  # rows being plotted from source
            try:
                k = tuple(c.getValue(row) for c in catcols) if catcols else (ycol.name,)

                # convert deliberately to float (to e.g. linearize date)
                graph_x = numcols[0].type(numcols[0].getValue(row)) if numcols else rownum
                graph_y = ycol.type(ycol.getValue(row))

                xpts[k].append(graph_x)
                ypts[k].append(graph_y)

                nplotted += 1
            except Exception:
                nerrors += 1
                if vd.options.debug:
                    raise
        lines = []
        for k in xpts:
            line = ax.scatter(xpts[k], ypts[k], label=' '.join(str(x) for x in k), **vd.options.getall('plt_'))
            lines.append(line)

        ax.legend(handles=lines)
        ax.set_xlabel(','.join(xcol.name for xcol in self.xcols if vd.isNumeric(xcol)) or 'row#')
        ax.xaxis.set_major_locator(plt.MaxNLocator(4))
        ax.yaxis.set_major_locator(plt.MaxNLocator(4))


@VisiData.api
def save_svg(vd, p, *sheets):
    plt = vd.importExternal('matplotlib.pyplot', 'matplotlib')
    fig_, ax = plt.subplots()
    for vs in sheets:
        if not isinstance(vs, Canvas):
            vd.warning(f'{vs.name} not a Canvas')
            continue
        vs.plot_sheet(ax)

    ax.grid()
    ax.set_title('\n'.join(vs.name for vs in sheets))
    plt.xticks()
    plt.savefig(p, format='svg')



import json

from visidata import vd, VisiData, Path, AttrDict


@VisiData.api
class StoredList(list):
    'Read existing persisted list from filesystem, and append new elements to .jsonl in .visidata'
    def __init__(self, *args, name:str='', **kwargs):
        super().__init__(*args, **kwargs)
        self.name = name

    @property
    def path(self):
        vdpath = Path(vd.options.visidata_dir)
        if not vdpath.exists():
            if vd.options.nothing:
                return
            vdpath.mkdir(parents=True)
        return vdpath/(self.name + '.jsonl')

    def reload(self):
        p = self.path
        if not p or not p.exists():
            return

        ret = []
        with p.open(encoding='utf-8-sig') as fp:
            for line in fp:
                value = vd.callNoExceptions(json.loads, line)
                if value is not None:
                    if isinstance(value, dict):
                        value = AttrDict(value)
                    ret.append(value)

        self[:] = ret   # replace without using .append

    def append(self, v):
        super().append(v)

        p = self.path
        if p is None:
            return

        with p.open(encoding='utf-8', mode='a') as fp:
            fp.write(json.dumps(v) + '\n')



import ctypes
import time
import os.path
import functools
import cProfile
import threading
import collections
import subprocess
import curses

from visidata import VisiData, vd, options, globalCommand, Sheet, EscapeException
from visidata import ColumnAttr, Column, BaseSheet, ItemColumn


vd.option('profile', False, 'enable profiling on threads')
vd.option('min_memory_mb', 0, 'minimum memory to continue loading and async processing')

vd.theme_option('color_working', '118 5', 'color of system running smoothly')

BaseSheet.init('currentThreads', list)

def asynccache(key=lambda *args, **kwargs: str(args)+str(kwargs)):
    def _decorator(func):
        'Function decorator, so first call to `func()` spawns a separate thread. Calls return the Thread until the wrapped function returns; subsequent calls return the cached return value.'
        d = {}  # per decoration cache
        def _func(k, *args, **kwargs):
            d[k] = func(*args, **kwargs)

        @functools.wraps(func)
        def _execAsync(*args, **kwargs):
            k = key(*args, **kwargs)
            if k not in d:
                d[k] = vd.execAsync(_func, k, *args, **kwargs)
            return d.get(k)
        return _execAsync
    return _decorator


class _Progress:
    def __init__(self, iterable=None, gerund="", total=None, sheet=None):
        self.iterable = iterable
        if total is None:
            if iterable is not None:
                self.total = len(iterable)
            else:
                self.total = 0
        else:
            self.total = total
        self.sheet = sheet if sheet else getattr(threading.current_thread(), 'sheet', None)
        self.gerund = gerund
        self.made = 0

    def __enter__(self):
        if self.sheet:
            self.sheet.progresses.insert(0, self)
        return self

    def addProgress(self, n):
        'Increase the progress count by *n*.'
        self.made += n
        return True

    def __exit__(self, exc_type, exc_val, tb):
        if self.sheet:
            self.sheet.progresses.remove(self)

    def __iter__(self):
        with self as prog:
            for item in self.iterable:
                yield item
                self.made += 1

@VisiData.global_api
def Progress(vd, iterable=None, gerund="", total=None, sheet=None):
    '''Maintain progress count as either an iterable wrapper, or a context manager.

        - *iterable*: wrapped iterable if used as an iterator.
        - *gerund*: status text shown while this Progress is active.
        - *total*: total count expected.
        - *sheet*: specific sheet to associate this progress with.  Default is sheet from current thread.
        '''
    return _Progress(iterable=iterable, gerund=gerund, total=total, sheet=sheet)


@VisiData.api
def cancelThread(vd, *threads, exception=EscapeException):
    'Raise *exception* in one or more *threads*.'
    for t in threads:
        if t.ident is not None:
            ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(t.ident), ctypes.py_object(exception))


# each row is an augmented threading.Thread object
class ThreadsSheet(Sheet):
    rowtype = 'threads'
    precious = False
    columns = [
        ColumnAttr('name'),
        Column('process_time', type=float, getter=lambda col,row: elapsed_s(row)),
        ColumnAttr('profile'),
        ColumnAttr('status'),
        ColumnAttr('exception'),
    ]
    def reload(self):
        self.rows = self.source

    def openRow(self, row):
        'push profile sheet for this action'
        if row.profile:
            return ProfileSheet(row.name, "profile", source=row.profile)
        vd.warning("no profile")


def elapsed_s(t):
    return (t.endTime or time.process_time())-t.startTime


@VisiData.api
def checkMemoryUsage(vd):
    threads = vd.unfinishedThreads
    if not threads:
        return ''

    min_mem = vd.options.min_memory_mb
    if not min_mem:
        return ''

    try:
        freestats = subprocess.run('free --total --mega'.split(), check=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE).stdout.strip().splitlines()
    except FileNotFoundError as e:
        if vd.options.debug:
            vd.exceptionCaught(e)
        vd.options.min_memory_mb = 0
        vd.warning('disabling min_memory_mb: "free" not installed')
        return ''
    tot_m, used_m, free_m = map(int, freestats[-1].split()[1:])
    ret = f'  [{free_m}MB] '
    if free_m < min_mem:
        attr = '[:warning]'
        vd.warning(f'{free_m}MB free < {min_mem}MB minimum, stopping threads')
        vd.cancelThread(*vd.unfinishedThreads)
        curses.flash()
    else:
        attr = '[:working]'
    return attr + ret + '[/]'


# for progress bar
BaseSheet.init('progresses', list)  # list of Progress objects

@BaseSheet.property
def progressMade(self):
    return sum(prog.made for prog in self.progresses if prog.total)

@BaseSheet.property
def progressTotal(self):
    return sum(prog.total for prog in self.progresses)


@BaseSheet.property
def progressPct(sheet):
    'Percent complete string as indicated by async actions.'
    if sheet.progresses and sheet.progressTotal > 0:
        return '%2d%%' % int(sheet.progressMade*100//sheet.progressTotal)
    return ''

## threads

def _annotate_thread(t, endTime=None):
    t.startTime = time.process_time()
    t.endTime = endTime  # endTime is None means unfinished.  endTime=0 for main thread
    t.status = ''
    t.profile = None
    t.exception = None
    return t

# all long-running threads, including main and finished
vd.threads = [_annotate_thread(threading.current_thread(), 0)]

@VisiData.api
def execSync(vd, func, *args, sheet=None, **kwargs):
    'Execute ``func(*args, **kwargs)`` in this thread (synchronously). A drop-in substitute for vd.execAsync.'
    vd.callNoExceptions(func, *args, **kwargs)
    t = threading.current_thread()
    t.sheet = sheet or vd.activeSheet
    return t

@VisiData.api
def execAsync(vd, func, *args, **kwargs):
    '''Execute ``func(*args, **kwargs)`` in a separate thread.  `sheet` is a
    special kwarg to indicate which sheet the thread should be associated with;
    by default, uses vd.activeSheet.  If `sheet` explicitly given as None, the thread
    will be ignored by vd.sync and thread status indicators.
    '''

    if 'sheet' not in kwargs:
        sheet = vd.activeSheet
    else:
        sheet = kwargs.pop('sheet')

    if sheet is not None and (sheet.lastCommandThreads and threading.current_thread() not in sheet.lastCommandThreads):
        vd.fail(f'still running **{sheet.lastCommandThreads[-1].name}** from previous command')

    # the current thread's activeCommand
    cmd = vd.activeCommand
    # the newly started thread is assigned the same activeCommand
    def _with_active_cmd(func, *args, **kwargs):
        if cmd: vd.activeCommand = cmd
        _toplevelTryFunc(func, *args, **kwargs)
        if cmd: vd.activeCommand = None
    thread = threading.Thread(target=_with_active_cmd, daemon=True, args=(func,)+args, kwargs=kwargs)
    vd.threads.append(_annotate_thread(thread))

    if sheet is not None:
        sheet.currentThreads.append(thread)

    thread.sheet = sheet
    thread.start()

    return thread

def _toplevelTryFunc(func, *args, **kwargs):
  with ThreadProfiler(threading.current_thread()) as prof:
    t = threading.current_thread()
    t.name = func.__name__
    try:
        t.status = func(*args, **kwargs)
    except EscapeException as e:  # user aborted
        t.status = 'aborted by user'
        vd.warning(f'{t.name} aborted')
    except Exception as e:
        t.exception = e
        t.status = 'exception'
        vd.exceptionCaught(e)

    if t.sheet:
        t.sheet.currentThreads.remove(t)

def asyncignore(func):
    'Decorator like `@asyncthread` but without attaching to a sheet, so no sheet.threadStatus will show it.'
    @functools.wraps(func)
    def _execAsync(*args, **kwargs):
        @functools.wraps(func)
        def _func(*args, **kwargs):
            func(*args, **kwargs)

        return vd.execAsync(_func, *args, **kwargs, sheet=None)

    return _execAsync

def asyncsingle(func):
    '''Function decorator like `@asyncthread` but as a singleton.  When called, `func(...)` spawns a new thread, and cancels any previous thread still running *func*.
    ``vd.sync()`` does not wait for unfinished asyncsingle threads.
    '''
    @functools.wraps(func)
    def _execAsync(*args, **kwargs):
        def _func(*args, **kwargs):
            func(*args, **kwargs)
            _execAsync.searchThread = None
            # end of thread

        # cancel previous thread if running
        if _execAsync.searchThread:
            vd.cancelThread(_execAsync.searchThread)

        _func.__name__ = func.__name__ # otherwise, the the thread's name is '_func'

        _execAsync.searchThread = vd.execAsync(_func, *args, **kwargs)
        _execAsync.searchThread.noblock = True
    _execAsync.searchThread = None
    return _execAsync

def asyncsingle_queue(func):
    '''Function decorator like `@asyncthread` but as a singleton.  When called, `func(...)` spawns a new thread, and waits for the end of any previous thread still running *func*.
    ``vd.sync()`` does wait for unfinished asyncsingle_queue threads, which is an important difference from asyncsingle.
    '''
    @functools.wraps(func)
    def _execAsync(*args, **kwargs):
        def _func(*args, **kwargs):
            func(*args, **kwargs)
            _execAsync.searchThread = None
            # end of thread

        # cancel previous thread if running
        if _execAsync.searchThread:
            vd.sync(_execAsync.searchThread)

        _func.__name__ = func.__name__ # otherwise, the the thread's name is '_func'

        _execAsync.searchThread = vd.execAsync(_func, *args, **kwargs)
    _execAsync.searchThread = None
    return _execAsync

@VisiData.property
def unfinishedThreads(self):
    'A list of unfinished threads (those without a recorded `endTime`).'
    return [t for t in self.threads if getattr(t, 'endTime', None) is None and getattr(t, 'sheet', None) is not None]

@VisiData.api
def checkForFinishedThreads(self):
    'Mark terminated threads with endTime.'
    for t in self.unfinishedThreads:
        if not t.is_alive():
            t.endTime = time.process_time()
            if getattr(t, 'status', None) is None:
                t.status = 'ended'

@VisiData.api
def sync(self, *joiningThreads):
    'Wait for one or more *joiningThreads* to finish. If no *joiningThreads* specified, wait for all but current thread and interface thread to finish.'
    joiningThreads = set(joiningThreads)
    while True:
        deads = set()  # dead threads
        threads = joiningThreads or set(self.unfinishedThreads)
        threads -= set([threading.current_thread(), getattr(vd, 'drawThread', None), getattr(vd, 'outputProgressThread', None)])
        threads -= deads
        threads -= set([None])
        for t in threads:
            try:
                if not t.is_alive() or t not in threading.enumerate() or getattr(t, 'noblock', False) is True:
                    deads.add(t)
                else:
                    t.join(timeout=1)
            except RuntimeError as e:  # maybe thread hasn't started yet or has already joined
                vd.exceptionCaught(e)
                pass

        if len(threads - deads) == 0:
            break


min_thread_time_s = 0.10 # only keep threads that take longer than this number of seconds

@VisiData.api
def open_pyprof(vd, p):
    import pstats
    return ProfileStatsSheet(p.base_stem, source=pstats.Stats(p.given).stats)


@VisiData.api
def toggleProfiling(vd):
    t = threading.current_thread()
    if not vd.options.profile:
        if not t.profile:
            t.profile = cProfile.Profile()
        t.profile.enable()
        vd.options.set('profile', True)
    else:
        if t.profile:
            t.profile.disable()
        vd.options.set('profile', False)
    vd.status('profiling ' + ('ON' if vd.options.profile else 'OFF'))


class ThreadProfiler:
    def __init__(self, thread):
        self.thread = thread
        self.thread.profile = cProfile.Profile()

    def __enter__(self):
        if vd.options.profile:
            try:
                self.thread.profile.enable()
            except ValueError: #"ValueError: Another profiling tool is already active"
                pass
        return self

    def __exit__(self, exc_type, exc_val, tb):
        self.thread.profile.disable()

        if exc_val:
            self.thread.exception = exc_val
        else:
            # remove very-short-lived async actions
            if elapsed_s(self.thread) < min_thread_time_s:
                vd.threads.remove(self.thread)
            else:
                if self.thread.sheet:
                    if vd.options.profile:
                        vd.status(f'[:bold]{self.thread.sheet.name[:32]}.{self.thread.name}[/] finished in {elapsed_s(self.thread):.1f}s')
                if vd.options.profile:
                    self.thread.profile.dump_stats(f'{self.thread.name}.pyprof')


class ProfileSheet(Sheet):
    rowtype = 'callsites' # rowdef: profiler_entry
    guide = '''
        # Profile Sheet
        - `z Ctrl+S` to save as pyprof file
        - `Ctrl+O` to open current function in $EDITOR
        - `Enter` to open list of calls from current function
    '''
    columns = [
        Column('funcname', getter=lambda col,row: codestr(row.code)),
        Column('filename', getter=lambda col,row: os.path.split(row.code.co_filename)[-1] if not isinstance(row.code, str) else ''),
        Column('linenum', type=int, getter=lambda col,row: row.code.co_firstlineno if not isinstance(row.code, str) else None),

        Column('inlinetime_us', type=int, getter=lambda col,row: row.inlinetime*1000000),
        Column('totaltime_us', type=int, getter=lambda col,row: row.totaltime*1000000),
        ColumnAttr('callcount', type=int),
        Column('avg_inline_us', type=int, getter=lambda col,row: row.inlinetime*1000000/row.callcount),
        Column('avg_total_us', type=int, getter=lambda col,row: row.totaltime*1000000/row.callcount),
        ColumnAttr('reccallcount', type=int),
        ColumnAttr('calls'),
        Column('callers', getter=lambda col,row: col.sheet.callers[row.code]),
    ]

    nKeys=3
    _ordering = [('inlinetime_us', True)]  # initially sort by inlinetime descending

    def reload(self):
        if isinstance(self.source, cProfile.Profile):
            self.rows = self.source.getstats()
        else:
            self.rows = self.source

        self.callers = collections.defaultdict(list)  # [row.code] -> list(code)

        for r in self.rows:
            calls = getattr(r, 'calls', None)
            if calls:
                for callee in calls:
                    self.callers[callee.code].append(r)

    def openRow(self, row):
        'open ProfileSheet for calls referenced in current row'
        if row.calls:
            return ProfileSheet(codestr(row.code)+"_calls", source=row.calls)
        vd.warning("no calls")

    def openCell(self, col, row):
        'open ProfileSheet for caller referenced in current cell'
        val = col.getValue(row)
        if val:
            return ProfileSheet(codestr(row.code)+"_"+col.name, source=val)
        vd.warning("no callers")


class ProfileStatsSheet(Sheet):
    rowtype = 'functions' # rowdef: list from pstats.Stats.stats
    columns = [
        ItemColumn('pathname', 0),
        ItemColumn('line', 1, type=int),
        ItemColumn('func', 2),
        ItemColumn('ncalls', 3, type=int),
        ItemColumn('primitive_calls', 4, type=int, width=0),
        ItemColumn('tottime', 5, type=float),
        ItemColumn('cumtime', 6, type=float),
        ItemColumn('callers', 7),
    ]
    def reload(self):
        self.rows = list((k+v) for k,v in self.source.items())

    def openRow(self, row):
        return ProfileStatsSheet('', source=row[7])

def codestr(code):
    if isinstance(code, str):
        return code
    return code.co_name


@VisiData.lazy_property
def allThreadsSheet(self):
    return ThreadsSheet("threads", source=vd.threads)

ThreadsSheet.addCommand('^C', 'cancel-thread', 'cancelThread(cursorRow)', 'abort thread at current row')
ThreadsSheet.addCommand('g^C', 'cancel-all', 'cancelThread(*sheet.rows)', 'abort all threads on this threads sheet')
ThreadsSheet.addCommand(None, 'add-row', 'fail("cannot add new rows on Threads Sheet")', 'invalid command')

ProfileSheet.addCommand('z^S', 'save-profile', 'source.dump_stats(input("save profile to: ", value=name+".prof"))', 'save profile')
ProfileSheet.addCommand('^O', 'sysopen-row', 'launchEditor(cursorRow.code.co_filename, "+%s" % cursorRow.code.co_firstlineno)', 'open current file at referenced row in external $EDITOR')
ProfileStatsSheet.addCommand('^O', 'sysopen-row', 'launchEditor(cursorRow[0], "+%s" % cursorRow[1])', 'open current file at referenced row in external $EDITOR')

BaseSheet.addCommand('^_', 'toggle-profile', 'toggleProfiling()', 'Enable or disable profiling on main VisiData process')

BaseSheet.addCommand('^C', 'cancel-sheet', 'cancelThread(*sheet.currentThreads or fail("no active threads on this sheet"))', 'abort all threads on current sheet')
BaseSheet.addCommand('g^C', 'cancel-all', 'liveThreads=list(t for vs in vd.sheets for t in vs.currentThreads); cancelThread(*liveThreads); status("canceled %s threads" % len(liveThreads))', 'abort all spawned threads')


BaseSheet.addCommand('^T', 'threads-all', 'vd.push(vd.allThreadsSheet)', 'open Threads for all sheets')
BaseSheet.addCommand('z^T', 'threads-sheet', 'vd.push(ThreadsSheet("threads", source=sheet.currentThreads))', 'open Threads for this sheet')

vd.addGlobals({
    'ThreadsSheet': ThreadsSheet,
    'Progress': Progress,
    'asynccache': asynccache,
    'asyncsingle': asyncsingle,
    'asyncsingle_queue': asyncsingle_queue,
    'asyncignore': asyncignore,
})

vd.addMenuItems('''
    System > Threads sheet > threads-all
    System > Toggle profiling > toggle-profile
''')



from collections import defaultdict

from visidata import VisiData, vd, Sheet, Column, asyncthread, CellColorizer, ColumnItem, ENTER


vd.option('color_xword_active', 'green', 'color of active clue')


@VisiData.api
def open_puz(vd, p):
    return PuzSheet(p.base_stem, source=p)

@VisiData.api
def open_xd(vd, p):
    if p.is_dir():
        return CrosswordsSheet(p.base_stem, source=p)
    return CrosswordSheet(p.base_stem, source=p)


@VisiData.api
class CrosswordsSheet(Sheet):
    rowtype = 'puzzles'
    columns = [
            Column('Author', getter=lambda col, row: row.author),
            Column('Copyright', getter=lambda col, row: row.copyright),
            Column('Notes', getter=lambda col, row: row.notes),
            Column('Postscript', getter=lambda col, row: ''.join(x for x in row.postscript if ord(x) >= ord(' '))),
            Column('Preamble', getter=lambda col, row: row.preamble),
            Column('Title', getter=lambda col, row: row.title)
            ]

    @asyncthread
    def reload(self):
        self.rows = []
        for p in self.source.iterdir():
            self.addRow(Crossword(p.read(), str(p)))


@VisiData.api
class GridSheet(Sheet):
    rowtype = 'gridrow'  # rowdef: puzzle_row:str
    colorizers = [
        CellColorizer(7, 'color_xword_active', lambda s,c,r,v: r and s.pos in s.cells[(s.rows.index(r),c)])
    ]

    @asyncthread
    def reload(self):
        grid = self.source.xd.grid

        ncols = len(grid[0])
        self.columns = [ColumnItem('', i, width=2) for i in range(ncols)]

        for row in grid:
            row = list(row)
            self.addRow(row)

        self.cells = defaultdict(list) # [rownum, col] -> [ Apos, Dpos ] or [] (if black)

        # find starting r,c from self.pos
        for cluedir, cluenum, answer, r, c in self.source.xd.iteranswers_full():
            # across
            if cluedir == 'A':
                for i in range(0, len(answer)):
                    self.cells[(r, self.columns[c+i])].append(('A', cluenum))
            if cluedir == 'D':
                for i in range(0, len(answer)):
                    self.cells[(r+i, self.columns[c])].append(('D', cluenum))

                if cluenum == self.pos[1]:
                    self.cursorRowIndex, self.cursorVisibleColIndex = r, c


class CrosswordSheet(Sheet):
    rowtype = 'clues' # rowdef: (cluenum, clue, answer)

    columns = [
            Column('clue_number', getter=lambda col, row: row[0][0]+str(row[0][1])),
            Column('clue', getter=lambda col, row: row[1]),
            Column('answer', getter=lambda col, row: row[2])
            ]

    @asyncthread
    def reload(self):
        import xdfile
        self.xd = xdfile.xdfile(xd_contents=self.source.read_text(), filename=self.source)
        self.rows = self.xd.clues


class PuzSheet(CrosswordSheet):
    @asyncthread
    def reload(self):
        import xdfile.puz2xd
        self.xd = xdfile.puz2xd.parse_puz(self.source.read_bytes(), str(self.source))
        self.rows = self.xd.clues


@VisiData.api
def save_xd(vd, p, vs):
    with p.open(mode='w', encoding='utf-8') as fp:
        fp.write(vs.xd.to_unicode())


CrosswordsSheet.addCommand(ENTER, 'open-clues', 'vd.push(CrosswordSheet("clues_"+cursorRow.title, source=cursorRow))', 'open CrosswordSheet: clue answer pair for crossword')
CrosswordSheet.addCommand(ENTER, 'open-grid', 'vd.push(GridSheet("grid", source=sheet, pos=cursorRow[0]))', 'open GridSheet: grid for crossword')

GridSheet.options.disp_column_sep = ''



from visidata import vd, BaseSheet

BaseSheet.addCommand('gO', 'open-config', 'vd.push(open_txt(Path(options.config)))', 'open options.config as text sheet')

vd.addMenuItems('''
    File > Options > edit config file > open-config
''')



from visidata import Sheet, asyncthread, Progress


@Sheet.api
@asyncthread
def select_equal_selected(sheet, col):
    selectedVals = set(col.getDisplayValue(row) for row in Progress(sheet.selectedRows))
    sheet.select(sheet.gatherBy(lambda r,c=col,vals=selectedVals: c.getDisplayValue(r) in vals), progress=False)


Sheet.addCommand('', 'select-equal-selected', 'select_equal_selected(cursorCol)', 'select rows with values in current column in already selected rows')



import curses
from visidata import VisiData, colors, Sheet, Column, ItemColumn, RowColorizer, wrapply, BaseSheet


class ColorSheet(Sheet):
    rowtype = 'colors'  # rowdef: [fg, bg, color_attr, colornamestr]
    columns = [
        ItemColumn('fg', 0, type=int),
        ItemColumn('bg', 1, type=int),
        ItemColumn('pairnum', 2),
        ItemColumn('name', 3),
    ]
    colorizers = [
        RowColorizer(7, None, lambda s,c,r,v: r and r[3])
    ]

    def iterload(self):
        self.rows = []
        for k, v in colors.color_pairs.items():
            fg, bg = k
            pairnum, colorname = v
            yield [fg, bg, pairnum, colorname]

        for i in range(0, 256):
            yield [i, 0, None, f'{i}']

    def draw(self, scr):
        super().draw(scr)
        rightcol = max(self._visibleColLayout.values())
        xstart = rightcol[0] + rightcol[1] + 4
        for i, r in enumerate(self.rows[(self.topRowIndex//6)*6:(self.bottomRowIndex//6+1)*6]):
            fg, bg, _, colorstr = r
            s = f'█▌{fg:3}▐█'
            y=i//6+1
            x=(i%6)*(len(s)+2)+xstart
            if y > self.windowHeight-1:
                break
            if r is self.cursorRow:
                s = f'█[{fg:3}]█'
            scr.addstr(y, x, s, colors[colorstr].attr)


BaseSheet.addCommand(None, 'open-colors', 'vd.push(vd.colorsSheet)', 'open Color Sheet with available terminal colors')

@VisiData.lazy_property
def colorsSheet(vd):
    return ColorSheet('colors')



import math


def bezier(x1, y1, x2, y2, x3, y3):
    'Generate (x,y) coordinates on quadratic curve from (x1,y1) to (x3,y3) with control point at (x2,y2).'
    yield (x1, y1)
    yield from _recursive_bezier(x1, y1, x2, y2, x3, y3)
    yield (x3, y3)


def _recursive_bezier(x1, y1, x2, y2, x3, y3, level=0):
    'from http://www.antigrain.com/research/adaptive_bezier/'
    m_approximation_scale = 10.0
    m_distance_tolerance = (0.5 / m_approximation_scale) ** 2
    m_angle_tolerance = 1 * 2*math.pi/360  # 15 degrees in rads
    curve_angle_tolerance_epsilon = 0.01
    curve_recursion_limit = 32
    curve_collinearity_epsilon = 1e-30

    if level > curve_recursion_limit:
        return

    # Calculate all the mid-points of the line segments

    x12   = (x1 + x2) / 2
    y12   = (y1 + y2) / 2
    x23   = (x2 + x3) / 2
    y23   = (y2 + y3) / 2
    x123  = (x12 + x23) / 2
    y123  = (y12 + y23) / 2

    dx = x3-x1
    dy = y3-y1
    d = abs(((x2 - x3) * dy - (y2 - y3) * dx))

    if d > curve_collinearity_epsilon:
        # Regular care
        if d*d <= m_distance_tolerance * (dx*dx + dy*dy):
            # If the curvature doesn't exceed the distance_tolerance value, we tend to finish subdivisions.
            if m_angle_tolerance < curve_angle_tolerance_epsilon:
                yield (x123, y123)
                return

            # Angle & Cusp Condition
            da = abs(math.atan2(y3 - y2, x3 - x2) - math.atan2(y2 - y1, x2 - x1))
            if da >= math.pi:
                da = 2*math.pi - da

            if da < m_angle_tolerance:
                # Finally we can stop the recursion
                yield (x123, y123)
                return
    else:
        # Collinear case
        dx = x123 - (x1 + x3) / 2
        dy = y123 - (y1 + y3) / 2
        if dx*dx + dy*dy <= m_distance_tolerance:
            yield (x123, y123)
            return

    # Continue subdivision
    yield from _recursive_bezier(x1, y1, x12, y12, x123, y123, level + 1)
    yield from _recursive_bezier(x123, y123, x23, y23, x3, y3, level + 1)



import functools
import collections

from visidata import VisiData, MetaSheet, ColumnAttr, Column, BaseSheet, VisiDataMetaSheet, SuspendCurses
from visidata import vd, asyncthread, ENTER, drawcache, AttrDict, TextSheet


vd.option('disp_help_flags', 'cmdpalette guides help hints inputfield inputkeys nometacols sidebar',
    '''list of helper features to enable (space-separated):
    - "cmdpalette": exec-longname suggestions
    - "guides": guides in sidebar
    - "help": help sidebar collapsed by default
    - "hints": context-sensitive hints on menu line
    - "inputfield": context-sensitive help for each input field
    - "inputkeys": input quick reference in sidebar
    - "nometacols": hide expert columns on metasheets
    - "sidebar": context-sensitive sheet help in sidebar
    - "all": enable all helper features''')


@VisiData.api
def wantsHelp(vd, feat):
    return feat in vd.options.disp_help_flags or 'all' in vd.options.disp_help_flags


@BaseSheet.api
def hint_basichelp(sheet):
    return 0, 'Alt+H to open the [:underline]H[/]elp menu'


@VisiData.api
def iterMenuPaths(vd, item=None, menupath=[]):
    'Generate (longname, menupath).'
    if item is None:
        item = vd.menus

    if isinstance(item, (list, tuple)):
        for m in item:
            yield from vd.iterMenuPaths(m, menupath)
    elif item.longname:
        yield item.longname, ' > '.join(menupath+[item.title])
    else:
        yield from vd.iterMenuPaths(item.menus, menupath+[item.title])


@VisiData.property
@drawcache
def menuPathsByLongname(vd):
    return dict(vd.iterMenuPaths())


@VisiData.api
class HelpSheet(MetaSheet):
    'Show all commands available to the source sheet.'
    rowtype = 'commands'
    precious = False
    _ordering = [('sheet', False), ('longname', False)]

    columns = [
        ColumnAttr('sheet'),
        ColumnAttr('module'),
        ColumnAttr('longname'),
        Column('menupath', width=0, cache=True, getter=lambda col,row: vd.menuPathsByLongname.get(row.longname, None)),
        Column('keystrokes', getter=lambda col,row: col.sheet.revbinds.get(row.longname, [None])[0]),
        Column('all_bindings', width=0, cache=True, getter=lambda col,row: list(set(col.sheet.revbinds.get(row.longname, [])))),
        Column('description', width=70, getter=lambda col,row: col.sheet.cmddict[(row.sheet, row.longname)].helpstr),
        ColumnAttr('execstr', width=0),
        Column('logged', width=0, getter=lambda col,row: vd.isLoggableCommand(row)),
    ]
    nKeys = 2

    def iterload(self):
        cmdlist = VisiDataMetaSheet('cmdlist', source=None)

        self.cmddict = {}
        if self.source:
            itcmds = vd.commands.iter(obj=self.source)
        else:
            itcmds = vd.commands.iterall()

        for (k, o), v in itcmds:
            yield v
            v.sheet = o
            self.cmddict[(v.sheet, v.longname)] = v

        for cmdrow in cmdlist.rows:
            k = (cmdrow.sheet, cmdrow.longname)
            if k in self.cmddict:
                self.cmddict[k].helpstr = cmdrow.helpstr

    @BaseSheet.lazy_property
    def revbinds(self):
        revbinds = collections.defaultdict(list)  # longname -> [keystrokes, ..]
        itbindings = vd.bindkeys.iterall()
        for (keystrokes, _), longname in itbindings:
            revbinds[longname].append(keystrokes)

        return revbinds


class HelpPane:
    def __init__(self, name):
        import visidata
        self.name = name
        self.scr = None
        self.parentscr = None
        self.amgr = visidata.AnimationMgr()

    @property
    def width(self):
        return self.amgr.maxWidth

    @property
    def height(self):
        return self.amgr.maxHeight

    def draw(self, scr, x=None, y=None, **kwargs):
        if not scr: return
#        if not vd.wantsHelp('statushelp'):
#            if self.scr:
#                self.scr.erase()
#                self.scr.refresh()
#                self.scr = None
#            return

        if y is None: y=0  # show at top of screen by default
        if x is None: x=0
        hneeded = self.amgr.maxHeight+3
        wneeded = self.amgr.maxWidth+4
        scrh, scrw = scr.getmaxyx()
        if not self.scr or scr is not self.parentscr:  # (re)allocate help pane scr
            if y >= 0:
                if y+hneeded < scrh:
                    yhelp = y+1
                else:
                    hneeded = max(0, min(hneeded, y-1))
                    yhelp = y-hneeded
            else:  # y<0
                yhelp = max(0, scrh-hneeded-1)

            if x >= 0:
                if x+wneeded < scrw:
                    xhelp = x+1
                else:
                    wneeded = max(0, min(wneeded, x-1))
                    xhelp = x-wneeded
            else:  # x<0
                xhelp = max(0, scrh-wneeded-1)

            self.scr = vd.subwindow(scr, xhelp, yhelp, wneeded, hneeded)
            self.parentscr = scr

        self.scr.erase()
        self.scr.box()
        self.amgr.draw(self.scr, y=1, x=2, **kwargs)
        self.scr.noutrefresh()


@VisiData.api
@functools.lru_cache(maxsize=None)
def getHelpPane(vd, name, module='visidata') -> HelpPane:
    ret = HelpPane(name)
    try:
        ret.amgr.load(name, (vd.pkg_resources_files(module)/f'ddw/{name}.ddw').open(encoding='utf-8'))
        ret.amgr.trigger(name, loop=True)
    except FileNotFoundError as e:
        vd.debug(str(e))
    except ModuleNotFoundError as e:
        vd.debug(str(e))
    except KeyError as e:
        vd.debug(str(e))
    return ret


@VisiData.api
def openManPage(vd):
    import os
    with SuspendCurses():
        module_path = vd.pkg_resources_files(__name__.split('.')[0])
        if os.system(' '.join(['man', str(module_path/'man/vd.1')])) != 0:
            vd.push(TextSheet('man_vd', source=module_path/'man/vd.txt'))


# in VisiData, g^H refers to the man page
BaseSheet.addCommand('g^H', 'sysopen-help', 'openManPage()', 'Show the UNIX man page for VisiData')
BaseSheet.addCommand('z^H', 'help-commands', 'vd.push(HelpSheet(name + "_commands", source=sheet, revbinds={}))', 'list commands and keybindings available on current sheet')
BaseSheet.addCommand('gz^H', 'help-commands-all', 'vd.push(HelpSheet("all_commands", source=None, revbinds={}))', 'list commands and keybindings for all sheet types')

BaseSheet.bindkey('KEY_F(1)', 'sysopen-help')
BaseSheet.bindkey('zKEY_F(1)', 'help-commands')
BaseSheet.bindkey('zKEY_BACKSPACE', 'help-commands')
BaseSheet.bindkey('gKEY_BACKSPACE', 'sysopen-help')

HelpSheet.addCommand(None, 'exec-command', 'quit(sheet); draw_all(); activeStack[0].execCommand(cursorRow.longname)', 'execute command on undersheet')
BaseSheet.addCommand(None, 'open-tutorial-visidata', 'launchBrowser("https://jsvine.github.io/intro-to-visidata/")', 'open https://jsvine.github.io/intro-to-visidata/')


vd.addGlobals(HelpSheet=HelpSheet)

vd.addMenuItems('''
    Help > VisiData tutorial > open-tutorial-visidata
    Help > All commands > help-commands-all
    Help > Quick reference > sysopen-help
    Help > Command list > help-commands
''')



'''
View sections in one or more .org files.  Easily edit sections within the system editor.

## Supported syntax

- `--- [comment]` at start of line starts a new section
- any number of `#` or `*` (followed by a space) leads a headline
- `#+key: value` adds metadata to next element (headline, table, section, )
- orgmode style tags: `:tag1:tag2:`
- `[[tagname]]` links to that set of tags
- `[[url]]` links to external url
- orgmode and markdown links
   - `[[url][linktext]]`
   - `[linktext](url)`
- all other markup/orgmode/whatever is ignored and passed through

## Usage

- `vd file.org`
- or `find orgfiles/ -name '*.org' | vd -f forg`
- or `vd orgfiles/ -f orgdir`
'''

import collections
import datetime
import os
import re

from visidata import vd, VisiData, Column, Sheet, ItemColumn, vlen, asyncthread, Path, AttrDict, date


@VisiData.api
def open_org(vd, p):
    return OrgSheet(p.base_stem, source=p, filetype='org')


@VisiData.api
def open_forg(vd, p):
    return OrgSheet(p.base_stem, source=p, filetype='forg')


@VisiData.api
def open_orgdir(vd, p):
    return OrgSheet(p.base_stem, source=p, filetype='orgdir')


def encode_date(dt=None):
    if not dt:
        dt = datetime.datetime.now()
    elif isinstance(dt, str):
        dt = datetime.datetime.fromisoformat(dt)

    s = '123456789abcdefghijklmnopqrstuvwxyz'
    return '%02d%s%s' % (dt.year % 100, s[dt.month-1], s[dt.day-1])


class OrgContentsColumn(Column):
    def setValue(self, row, v, setModified=True):
        super().setValue(row, v, setModified=setModified)
        orgmode_parse_into(row, v)

    def putValue(self, row, v):
        self.sheet.save(row)


def sectionize(lines):
    'Generate (startinglinenum, contentlines) for each section.  First section may not have leading * or # but all others will.'
    startinglinenum = 0
    prev_contents = contents = []
    for linenum, line in enumerate(lines):
        line = line.rstrip('\n')

        if line and line[0] in '#*':
            if contents:
                yield startinglinenum, contents
                prev_contents = contents
                contents = []
                startinglinenum = linenum
        else:
            if not line and not contents:
                prev_contents.append(line)

        contents.append(line)

    if contents:
        yield startinglinenum, contents


def orgmode_parse(all_lines):
    root = parent = OrgSheet().newRow()
    root.orig_contents = ''
    for linenum, lines in sectionize(all_lines):
        section = OrgSheet().newRow()

        section.contents = ''
        for i, line in enumerate(lines):
            if not line and not lines[i-1]:
                continue
            section.contents += line + '\n'
        section.orig_contents = section.contents

        section.linenum = linenum+1
        for line in lines:
            section.tags.extend(re.findall(r':([\S:]+?):', line))
            links = re.findall(r'\[.*?\]\(.*?\)', line)
            if links:
                section.links.extend(links)

        title = orgmode_parse_title(lines[0])
        if not title:
            root = parent = section
            continue

        section.update(title)

        while parent and section.level <= parent.level:
            parent = parent.parent

        parent.children.append(section)
        section.parent = parent
        parent = section

    return root

def _replace(node, newnode):
    node.update(newnode)  # must replace insides of same row object
    for c in node.children:
        c.parent = node
    return node

def orgmode_parse_into(toprow, text):
    row = orgmode_parse(text.splitlines())
    if not row.title and len(row.children) == 1:
        row = row.children[0]
        row.parent = toprow.parent
    toprow = _replace(toprow, row)

    return toprow


def orgmode_to_string(section, prestars=''):
    ret = ''

#    if section.title:
#        ret += prestars[len(section.stars):] + (section.title or ' ') + '\n'
    ret += section.contents.rstrip() or ''
    ret += '\n\n'
    ret += ''.join(orgmode_to_string(c, prestars+(section.stars or '')) for c in section.children).rstrip() + '\n'

    return ret


def orgmode_parse_title(line):
    m = re.match(r'^(?P<stars>[*#]+)\s*(?P<keyword>(TODO|FEEDBACK|VERIFY|DONE|DELEGATED))?\s*(?P<prio>\[#[A-z]\])?\s*(?P<title>.*)', line)
    if not m:
        assert not line or line[0] not in '#*', line
        return {}

    return dict(stars=m.group('stars'),
                level=len(m.group('stars')),
                keyword=m.group('keyword') or '',
                prio=m.group('prio') or '',
                title=line)


class OrgSheet(Sheet):
    guide = '''# Orgmode Sheet (experimental)
A list of orgmode sections from _{sheet.source}_.

- `Enter` to expand current section
- `z Enter` to contract current section
- `ga` to combine selected sections into a new entry
- `Ctrl+O` to edit section in system editor (edits source directly)
- `g Ctrl+S` to save all org files
'''
    defer = True
    columns = [
        Column('path', getter=lambda c,r: _root(r).path, width=0),
        Column('id', getter=lambda c,r: _root(r).path.stem),
        ItemColumn('title', width=40),
#        ItemColumn('date', width=0, type=date),
        ItemColumn('tags', width=10, type=lambda v: ' '.join(v)),
        ItemColumn('links', type=vlen),
#        ItemColumn('parent', width=0),
        ItemColumn('children', type=vlen),
        ItemColumn('linenum', width=0, type=int),
        Column('to_string', width=0, getter=lambda c,r: orgmode_to_string(r)),
        OrgContentsColumn('orig_contents', width=0, getter=lambda c,r: r.orig_contents),
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.opened_rows = set()
#        self.open_max=0  # all levels closed after this point
#        self.close_max=0 # all levels open after this point

    def isSelectedParents(self, row):
        return super().isSelected(row) or row.parent and self.isSelectedParents(row.parent)

    def isSelected(self, row):
        return self.isSelectedParents(row)

    def refreshRows(self):
        self.rows = list(self._deepiter(self.sourceRows))

    def _deepiter(self, objlist, depth=1):
        for obj in objlist:
            if not obj.parent and obj.children:
                # ignore toplevel file, dive into subrows
                yield from self._deepiter(obj.children, depth)
            else:
                yield obj
                if id(obj) in self.opened_rows:
                    yield from self._deepiter(obj.children, depth-1)

    def openRows(self, rows):
        for row in rows:
            self.opened_rows.add(id(row))
        self.refreshRows()

    def closeRows(self, rows=None):
        if rows is None:
            for row in rows:
                self.opened_rows.remove(id(row))
        else:
            self.opened_rows.clear()
        self.refreshRows()

    def newRow(self):
        return AttrDict(title='', contents='', tags=[], children=[], links=[], level=0, linenum=0)

    def iterload(self):
        self.rows = []
        def _walkfiles(p):
            basepath = str(p)
            for folder, subdirs, files in os.walk(basepath):
                subfolder = folder[len(basepath)+1:]
                if subfolder.startswith('.'): continue
                if subfolder in ['.', '..']: continue

                fpath = Path(folder)
                yield fpath

                for fn in files:
                    yield fpath/fn

        if self.filetype == 'orgdir':
            basepath = str(self.source)
            for p in _walkfiles(self.source):
                if p.base_stem.startswith('.'): continue
                if p.ext in ['org', 'md']:
                    yield self.parse_orgmd(p)
        elif self.filetype == 'forg':
            for fn in self.source.open():
                yield self.parse_orgmd(Path(fn.rstrip()))
        elif self.filetype == 'org':
            yield self.parse_orgmd(self.source)

        self.sourceRows = self.rows
        self.refreshRows()

    def parse_orgmd(self, path):
#        row.file_string = open(path).read()
        row = orgmode_parse(open(path).readlines())
        st = path.stat()
        if st:
            mtime = st.st_mtime

        row.path = path
        row.date = mtime
        return row

    def draw(self, scr):
        super().draw(scr)

    @asyncthread
    def putChanges(self):
        adds, mods, dels = self.getDeferredChanges()

        saveset = {}  # path:_root(row)
        for r in adds.values():
            saveset[_root(r).path] = _root(r)
        for r, _ in mods.values():
            saveset[_root(r).path] = _root(r)

        for row in saveset.values():
            self.save(row)

        self.commitAdds()

        self.commitMods()

        for row in dels.values():
            row.path.rename('%s-%s' % (row.path, encode_date()))

        self.commitDeletes()

    def save_all(self):
        for row in self.sourceRows:
            self.save(row)
        vd.status('saved %s org files' % len(self.sourceRows))

    def save(self, row):
        try:
            row.path.rename('%s-%s' % (row.path, encode_date()))  # backup
        except FileNotFoundError:
            pass

        with row.path.open(mode='w') as f:
            print(orgmode_to_string(row).rstrip(), file=f)


@OrgSheet.api
def paste_into(sheet, row, sourcerows, cols):
    row.children.extend(sourcerows)
    for r in sourcerows:
        r.parent.children.remove(r)
        r.parent = row
    sheet.refreshRows()


@OrgSheet.api
def paste_data_into(sheet, row, sourcerows, cols):
    body = row.body or ''
    for r in sourcerows:
        data = vd.encode_json(r, cols)
        row.contents += f':{cols[0].sheet.name}:{data}\n'


@OrgSheet.api
def combine_rows(sheet, rows):
    newrow = sheet.newRow()
    newrow.date = datetime.datetime.today()
    orgid = vd.cleanName(rows[0].title)
    newrow.path = Path((orgid or encode_date())+'.org')
    for r in rows:
        hdr = sheet.newRow()
        if hdr.title:
            hdr.update(orgmode_parse_title(hdr.title))
        hdr.children = list(r.children)
        newrow.children.append(hdr)
#    newrow.title = ' '
    newrow.file_string = orgmode_to_string(newrow, '*')
    return newrow


def _root(row):
    while row and row.parent:
        row = row.parent
    return row


@OrgSheet.api
def sysopen_row(sheet, row):
    root = _root(row)
    if root.path.exists():
        vd.launchEditor(root.path, '+%s'%row['linenum'])
    else:
        orgmode_parse_into(row, vd.launchExternalEditor(root.file_string))

@VisiData.api
def save_org(vd, p, *vsheets):
    with p.open(mode='w', encoding=vsheets[0].options.save_encoding) as fp:
        for vs in vsheets:
            if isinstance(vs, OrgSheet):
                for row in vs.rows:
                    print(orgmode_to_string(row).strip(), file=fp)
            else:
                vd.warning('not implemented')


@OrgSheet.api
def sysopen_rows(sheet, rows):
    ret = ''
    for r in rows:
        s = orgmode_to_string(r).strip()
        ret += '{::%s::}\n%s\n\n' % (id(r), s)

    ret = vd.launchExternalEditor(ret + "{::}")

    idrows = {id(r):r for r in rows}

    idtexts = {}
    for idtag, text in re.findall(r'\{::(\d+)::\}(.*?)(?={::)', ret, re.DOTALL):
        idtexts[int(idtag)] = text

    # reparse all existing
    for rowid, row in idrows.items():
        text = idtexts.get(rowid, '').strip()
        if text:
            orgmode_parse_into(row, text)
        else:
            # sometimes, top levels are missing parents
            # this needs to be debugged
            if row.parent:
                row.parent.children.remove(row)

    lastrow = None
    # find new rows to add
    for rowid, text in idtexts.items():
        if rowid not in idrows:
            section = orgmode_parse(text.splitlines())
            while section.level <= lastrow.level:
                lastrow = lastrow.parent

            if lastrow:
                sourceRows.append(section)
            else:
                sheet.addRow(section)

    sheet.refreshRows()

OrgSheet.addCommand('^O', 'sysopen-row', 'sysopen_row(cursorRow)', 'open current file in external $EDITOR')
OrgSheet.addCommand('g^O', 'sysopen-rows', 'sysopen_rows(selectedRows)', 'open selected files in external $EDITOR')
OrgSheet.addCommand('^J', 'expand-row', 'openRows([cursorRow]); sheet.cursorRowIndex += 1')
OrgSheet.addCommand('z^J', 'close-row', 'closeRows([cursorRow]); sheet.cursorRowIndex += 1')
OrgSheet.addCommand('g^J', 'expand-selected', 'openRows(selectedRows)')
OrgSheet.addCommand('gz^J', 'close-selected', 'closeRows(selectedRows)')
OrgSheet.addCommand('ga', 'combine-selected', 'addRows([combine_rows(selectedRows)], index=cursorRowIndex); cursorDown(1)', 'combine selected rows into new org entry')

OrgSheet.addCommand('zp', 'paste-data', 'paste_data_into(cursorRow, vd.memory.cliprows, vd.memory.clipcols)', 'move clipboard rows to children of current row')
OrgSheet.addCommand('p', 'paste-sections', 'paste_data_into(cursorRow, vd.memory.cliprows, vd.memory.clipcols)', 'move clipboard rows to children of current row')
OrgSheet.addCommand('g^S', 'save-all', 'save_all()', 'save all org files')


if __name__ == '__main__':
    for fn in sys.argv[1:]:
        sect = orgmode_parse(Path(fn))
        print(orgmode_to_string(sect).strip())



import re

from visidata import Path, RepeatFile, vd, VisiData
from visidata.loaders.tsv import splitter

vd.option('http_max_next', 0, 'max next.url pages to follow in http response') #848
vd.option('http_req_headers', {}, 'http headers to send to requests')
vd.option('http_ssl_verify', True, 'verify host and certificates for https')


@VisiData.api
def guessurl_mimetype(vd, path, response):
    content_filetypes = {
        'tab-separated-values': 'tsv'
    }

    for k in dir(vd):
        if k.startswith('open_'):
            ft = k[5:]
            content_filetypes[ft] = ft

    contenttype = response.getheader('content-type')
    subtype = contenttype.split(';')[0].split('/')[-1]
    if subtype in content_filetypes:
        return dict(filetype=content_filetypes.get(subtype), _likelihood=10)



@VisiData.api
def openurl_http(vd, path, filetype=None):
    schemes = path.scheme.split('+')
    if len(schemes) > 1:
        sch = schemes[0]
        openfunc = getattr(vd, f'openhttp_{sch}', vd.getGlobals().get(f'openhttp_{sch}'))
        if not openfunc:
            vd.fail(f'no vd.openhttp_{sch}')
        return openfunc(Path(schemes[-1]+'://'+path.given.split('://')[1]))

    import urllib.request
    import urllib.error
    import mimetypes

    ctx = None
    if not vd.options.http_ssl_verify:
        import ssl

        ctx = ssl.create_default_context()
        ctx.check_hostname = False
        ctx.verify_mode = ssl.CERT_NONE

    req = urllib.request.Request(path.given, **vd.options.getall('http_req_'))
    try:
        response = urllib.request.urlopen(req, context=ctx)
    except urllib.error.HTTPError as e:
        vd.fail(f'cannot open URL: HTTP Error {e.code}: {e.reason}')
    except urllib.error.URLError as e:
        vd.fail(f'cannot open URL: {e.reason}')

    filetype = filetype or vd.guessFiletype(path, response, funcprefix='guessurl_').get('filetype')  # try guessing by url
    filetype = filetype or vd.guessFiletype(path, funcprefix='guess_').get('filetype')  # try guessing by contents

    # Automatically paginate if a 'next' URL is given
    def _iter_lines(path=path, response=response, max_next=vd.options.http_max_next):
        path.responses = []
        n = 0
        while response:
            path.responses.append(response)
            with response as fp:
                for line in splitter(response, delim=b'\n'):
                    yield line.decode(vd.options.encoding)

            linkhdr = response.getheader('Link')
            src = None
            if linkhdr:
                links = parse_header_links(linkhdr)
                link_data = {}
                for link in links:
                    key = link.get('rel') or link.get('url')
                    link_data[key] = link
                src = link_data.get('next', {}).get('url', None)

            if not src:
                break

            n += 1
            if n > max_next:
                vd.warning(f'stopping at max next pages: {max_next} pages')
                break

            vd.status(f'fetching next page from {src}')
            req = urllib.request.Request(src, **vd.options.getall('http_req_'))
            response = urllib.request.urlopen(req)

    # add resettable iterator over contents as an already-open fp
    path.fptext = RepeatFile(_iter_lines())

    return vd.openSource(path, filetype=filetype)

def parse_header_links(link_header):
    '''Return a list of dictionaries:
    [{'url': 'https://example.com/content?page=1', 'rel': 'prev'},
     {'url': 'https://example.com/content?page=3', 'rel': 'next'}]
    Takes a link header string, of the form
    '<https://example.com/content?page=1>; rel="prev", <https://example.com/content?page=3>; rel="next"'
    See https://datatracker.ietf.org/doc/html/rfc8288#section-3
    '''

    links = []
    quote_space = ' \'"'
    link_header = link_header.strip(quote_space)
    if not link_header: return []
    for link_value in re.split(', *<', link_header):
        if ';' in link_value:
            url, params = link_value.split(';', maxsplit=1)
        else:
            url, params = link_value, ''
        link = {'url': url.strip('<>' + quote_space)}

        for param in params.split(';'):
            if '=' in param:
                key, value = param.split('=')
                key = key.strip(quote_space)
                value = value.strip(quote_space)
                link[key] = value
            else:
                break
        links.append(link)
    return links

VisiData.openurl_https = VisiData.openurl_http
